import json

import requests
from django.utils.translation import ugettext_lazy as _

from . import base

__all__ = ('Backend',)


class Backend(base.Backend):

    ID = 'slack-webhook'
    PROTOCOL = 'slack-webhook'
    ESCAPE_HTML = False
    EDITOR_TYPE = 'basic'
    VERBOSE_NAME = _('Slack')

    def send_message(self, _from_address, to_addresses, message):
        data = json.dumps({'text': message})
        headers = {'Content-Type': 'application/json'}
        for address in to_addresses:
            result = requests.post(address, data=data, headers=headers)
            if not result.ok:
                raise requests.HTTPError(result.text)



import collections
import pydoc

import django.urls.base
import django.urls.resolvers
import django.utils.encoding
import django.utils.regex_helper
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

from django_vox import settings

PROTOCOLS = {
    'email': _('Email'),
    'sms': _('SMS'),
    'slack-webhook': _('Slack Webhook'),
    'json-webhook': _('JSON Webhook'),
    'twitter': _('Twitter'),
    'xmpp': _('XMPP'),
    'activity': _('Activity Stream'),
}

PREFIX_NAMES = {
    'si': _('Site Contacts'),
    'c': '__content__',
    'se': _('Actor'),
    're': _('Target'),
}
PREFIX_FORMATS = {
    'c': '{}',
    'se': _('Actor\'s {}'),
    're': _('Target\'s {}'),
}

_CHANNEL_TYPE_IDS = None


class ObjectNotFound(Exception):
    pass


class BackendManager:

    def __init__(self, class_list):
        self.proto_map = collections.defaultdict(list)
        self.id_map = {}
        for cls in class_list:
            if cls.ID in self.id_map:
                raise RuntimeError(
                    'Conflicting backend IDs: {}'.format(cls.ID))
            self.proto_map[cls.PROTOCOL].append(cls)
            self.id_map[cls.ID] = cls

    def by_protocol(self, protocol: str):
        return self.proto_map[protocol]

    def by_id(self, id_val):
        return self.id_map[id_val]

    def all(self):
        return self.id_map.values()

    def protocols(self):
        return self.proto_map.keys()


UnboundChannel = collections.namedtuple(
    'UnboundChannel', ('name', 'target_class', 'func'))


class Channel:
    def __init__(self, ubc: UnboundChannel, obj):
        self.name = ubc.name
        self.target_class = ubc.target_class
        self.func = ubc.func
        self.obj = obj

    def contactables(self):
        return (self.obj,) if self.func is None else self.func(self.obj)


class UnboundChannelMap(dict):

    def bind(self, obj):
        return BoundChannelMap(
            ((key, Channel(ubc, obj)) for (key, ubc) in self.items()))


class BoundChannelMap(dict):
    pass


class ChannelManagerItem:

    def __init__(self, cls):
        self.cls = cls
        self.__prefixes = {}
        self._channels = collections.defaultdict(dict)

    def __bool__(self):
        return bool(self._channels)

    def add(self, key, name, target_type, func):
        self._channels[key] = name, target_type, func
        self.__prefixes = {}

    def add_self(self):
        self.add('', '', self.cls, None)

    def prefix(self, prefix) -> UnboundChannelMap:
        # get channels by prefix
        if prefix not in self.__prefixes:
            ubc_map = UnboundChannelMap()
            for key, (name, cls, func) in self._channels.items():
                channel_key = prefix if key == '' else prefix + ':' + key
                if name == '':
                    name = PREFIX_NAMES[prefix]
                    if name == '__content__':
                        name = self.cls._meta.verbose_name.title()
                else:
                    name = PREFIX_FORMATS[prefix].format(name)
                ubc_map[channel_key] = UnboundChannel(name, cls, func)
            self.__prefixes[prefix] = ubc_map
        return self.__prefixes[prefix]


class ChannelProxyManager(dict):

    def __missing__(self, key):
        if key not in objects:
            objects.add(key, regex=None)
        return objects[key].channels


class ObjectManagerItem:

    def __init__(self, cls):
        self.cls = cls
        self.pattern = None
        self.matcher = None
        self.reverse_form = None
        self.reverse_params = ()
        self.channels = ChannelManagerItem(cls)

    @property
    def has_url(self):
        return self.matcher is not None

    def set_regex(self, pattern: str):
        self.pattern = pattern
        if hasattr(django.urls.resolvers, 'RegexPattern'):
            self.matcher = django.urls.resolvers.RegexPattern(pattern)
        else:
            self.matcher = django.urls.resolvers.RegexURLPattern(pattern, None)
        normal = django.utils.regex_helper.normalize(pattern)
        self.reverse_form, self.reverse_params = next(iter(normal), ('', ()))

    def match(self, path: str):
        if self.matcher is None:
            return False
        if hasattr(django.urls.resolvers, 'RegexPattern'):
            return self.matcher.match(path)
        return self.matcher.resolve(path)

    def reverse(self, obj):
        if self.reverse_form is None:
            return None
        kwargs = dict(((param, getattr(obj, param, None))
                      for param in self.reverse_params))
        return '/' + (self.reverse_form % kwargs)


class ObjectManager(dict):

    def __missing__(self, key):
        item = ObjectManagerItem(key)
        self[key] = item
        return item

    def __setitem__(self, key, value: ObjectManagerItem):
        super().__setitem__(key, value)

    def __getitem__(self, key) -> ObjectManagerItem:
        return super().__getitem__(key)

    def add(self, cls, *, regex=...):
        if regex is ...:
            raise RuntimeError(
                'Must set regex keyword argument, '
                'use None if object has no URL')
        item = ObjectManagerItem(cls)
        if regex is not None:
            item.set_regex(regex)
        self[cls] = item

    def create_local_object(self, path):
        for key, value in self.items():
            match = value.match(path)
            if match:
                # RegexURLPattern vs RegexPattern
                kwargs = match.kwargs if hasattr(match, 'kwargs') else match[2]
                return key.__class__(**kwargs)
        msg = _('Unable to find class for {}.').format(path)
        raise ObjectNotFound(msg)

    def get_local_object(self, path):
        matched_patterns = []
        for key, value in self.items():
            match = value.match(path)
            if match:
                matched_patterns.append(value.pattern)
                # RegexURLPattern vs RegexPattern
                kwargs = match.kwargs if hasattr(match, 'kwargs') else match[2]
                try:
                    return key.objects.get(**kwargs)
                except key.DoesNotExist:
                    pass
        msg = _('Unable to find object for {}.').format(path)
        if matched_patterns:
            msg = msg + ' ' + _(
                'We tried the following patterns: {}.').format(
                ', '.join(matched_patterns))
        raise ObjectNotFound(msg)


backends = BackendManager(
    pydoc.locate(name) for name in settings.BACKENDS)

channels = ChannelProxyManager()

objects = ObjectManager()


def get_protocol_choices():
    for protocol in backends.protocols():
        yield (protocol, PROTOCOLS.get(protocol, protocol))


def _channel_type_ids():
    for all_models in apps.all_models.values():
        for model in all_models.values():
            if model in objects:
                if objects[model].channels:
                    ct = ContentType.objects.get_for_model(model)
                    yield ct.id


def channel_type_limit():
    global _CHANNEL_TYPE_IDS
    if _CHANNEL_TYPE_IDS is None:
        _CHANNEL_TYPE_IDS = tuple(_channel_type_ids())
    return {'id__in': _CHANNEL_TYPE_IDS}



# -*- coding: utf-8 -*-

import django.urls


def activity_inbox_middleware(get_response):
    def middleware(request):
        # setting request.urlconf doesn't work because that screws up
        # calls to reverse() within the normal django app
        if 'application/activity+json' in request.META.get('HTTP_ACCEPT', ''):
            urlconf = 'django_vox.activity_urls'
            resolver = django.urls.get_resolver(urlconf)
            resolver_match = resolver.resolve(request.path_info)
            callback, callback_args, callback_kwargs = resolver_match
            request.resolver_match = resolver_match

            return callback(request, *callback_args, **callback_kwargs)

        return get_response(request)

    return middleware



"""
URL patterns for testing app.

We don't have any for now.
"""
from django.conf.urls import include, url
from django.contrib import admin
from django.contrib.staticfiles.views import serve

from . import urls

urlpatterns = [
    url('^admin/', admin.site.urls),
    url(r'^', include(urls)),
] + [
    url(r'^static/(?P<path>.*)$', serve),
]



"""
Django settings for django vox demo
"""

from tests.settings import *  # noqa: F403,F401
from tests.settings import BASE_INSTALLED_APPS

INSTALLED_APPS = BASE_INSTALLED_APPS + [
    'tests.apps.VoxDemoConfig']

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

import tests.models
from django_vox.management.commands import make_notifications
from django_vox.models import Notification, Template


class MakeNotificationTests(TestCase):

    @staticmethod
    def test_kill_orphans():
        """If you completely remove a class,
        its notification should get deleted."""
        cmd = make_notifications.Command()
        cmd.handle(verbosity=0)
        assert 3 == Notification.objects.all().count()
        # simulate a deleted class
        fake_ct = ContentType.objects.create(
            app_label='blarg', model='deleted')
        Notification.objects.create(
            codename='foo', object_type=fake_ct, from_code=True)
        assert 4 == Notification.objects.all().count()
        cmd.handle(verbosity=0)
        assert 3 == Notification.objects.all().count()

    @staticmethod
    def test_keep_orphans_with_templates():
        cmd = make_notifications.Command()
        cmd.handle(verbosity=0)
        assert 3 == Notification.objects.all().count()
        # simulate a deleted class
        fake_ct = ContentType.objects.create(
            app_label='django_vox', model='deleted')
        notification = Notification.objects.create(
            codename='foo', object_type=fake_ct, from_code=True)
        Template.objects.create(
            notification=notification, subject='subject', content='content')
        cmd.handle(verbosity=0)

    @staticmethod
    def test_keep():
        """Make sure we keep notifications with the same IDs
        even if they have no templates"""
        cmd = make_notifications.Command()
        cmd.handle(verbosity=0)
        ids = set(v['id'] for v in Notification.objects.values('id'))
        assert 3 == len(ids)
        notification = Notification.objects.all()[0]
        Template.objects.filter(notification=notification).delete()
        cmd.handle(verbosity=0)
        second_ids = set(v['id'] for v in Notification.objects.values('id'))
        assert ids == second_ids

    @staticmethod
    def test_make_notifications():
        cmd = make_notifications.Command()
        assert 0 == Notification.objects.all().count()
        # dry run
        cmd.handle(dry_run=True)
        assert 0 == Notification.objects.all().count()
        # test basic notification making
        cmd.handle(verbosity=0)
        assert 3 == Notification.objects.all().count()
        # gather some general things
        article_ct = ContentType.objects.get_for_model(
            tests.models.Article)
        user_ct = ContentType.objects.get_for_model(
            tests.models.User)

        # make a notification to delete
        Notification.objects.create(
            codename='foo',
            object_type=article_ct,
            from_code=True,
        )
        assert Notification.objects.all().count() == 4
        cmd.handle()
        assert Notification.objects.all().count() == 3
        # here's one that shouldn't get deleted
        Notification.objects.create(
            codename='foo',
            object_type=article_ct,
            from_code=False,
        )
        assert Notification.objects.all().count() == 4
        cmd.handle()
        assert Notification.objects.all().count() == 4
        # now we'll change one and it should get reverted
        acn = Notification.objects.get_by_natural_key(
            'tests', 'article', 'create')
        acn.object_model = user_ct
        acn.save()
        cmd.handle(verbosity=0)
        acn = Notification.objects.get_by_natural_key(
            'tests', 'article', 'create')
        assert acn.actor_type is not None



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-06-22 16:17
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_vox', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TemplateAttachment',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('key', models.CharField(
                    max_length=500, verbose_name='key')),
                ('template', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='django_vox.Template', verbose_name='template')),
            ],
            options={
                'verbose_name': 'template attachment',
            },
        ),
        migrations.AlterField(
            model_name='sitecontact',
            name='address',
            field=models.CharField(
                blank=True, max_length=500, verbose_name='address'),
        ),
    ]



import json
from typing import List
from urllib.parse import urlsplit

import aspy
import django.db.transaction
import django.utils.encoding
from django.template import Context
from django.utils.translation import ugettext_lazy as _

import django_vox.base
import django_vox.models
import django_vox.registry

from . import base

__ALL__ = ('make_contact', 'Backend',)

PUBLIC_ADDRESS = 'https://www.w3.org/ns/activitystreams#Public'


def id_to_user(ids: List[str]):
    for identifier in ids:
        url = django.utils.encoding.iri_to_uri(identifier)
        # I'm pretty sure we're only path matching, but this might change
        path = urlsplit(url)[2]
        # there should be a '/' in the beginning of the path, but that's
        # not normally part of the regex
        if path.startswith('/'):
            path = path[1:]
        # put the message in the relevant inbox
        yield django_vox.registry.objects.get_local_object(path)


class Backend(base.Backend):

    ID = 'activity'
    PROTOCOL = 'activity'
    VERBOSE_NAME = _('Activity Stream')
    USE_SUBJECT = True
    EDITOR_TYPE = 'html-light'

    # aspy.Create is the default, don't need to use it
    ACTIVITIES = (
        aspy.Accept, aspy.TentativeAccept, aspy.Add, aspy.Delete, aspy.Follow,
        aspy.Ignore, aspy.Join, aspy.Leave, aspy.Like, aspy.Offer, aspy.Invite,
        aspy.Reject, aspy.TentativeReject, aspy.Remove, aspy.Undo, aspy.Update,
        aspy.View, aspy.Listen, aspy.Read, aspy.Move, aspy.Travel,
        aspy.Announce, aspy.Block, aspy.Flag, aspy.Dislike, aspy.Question)
    ACTIVITY_MAP = dict((activity.__name__.lower(), activity)
                        for activity in ACTIVITIES)

    @classmethod
    def _build_summary(cls, body: str, context):
        template = base.template_from_string(body)
        return template.render(context)

    @classmethod
    def _build_subject(cls, subject: str, context):
        if subject == '':
            obj = context.get('object')
            if obj is not None:
                if hasattr(obj, 'name'):
                    if isinstance(obj.name, str):
                        return obj.name
            return ''

        template = base.template_from_string(subject)
        return template.render(context)

    @classmethod
    def _build(cls, subject: str, body: str, parameters: dict):
        context = Context(parameters, autoescape=cls.ESCAPE_HTML)
        return (cls._build_subject(subject, context),
                cls._build_summary(body, context))

    @staticmethod
    def to_activity_object(obj):
        if isinstance(obj, (aspy.Object, aspy.Link)):
            return obj
        if hasattr(obj, '__activity__'):
            return obj.__activity__()
        if hasattr(obj, 'get_absolute_url'):
            return django_vox.base.full_iri(obj.get_absolute_url())
        return None

    @classmethod
    def build_message(cls, subject: str, body: str, parameters: dict,
                      _attachments: List[base.AttachmentData]):
        subject, summary = cls._build(subject, body, parameters)
        activity_type = parameters.get('activity_type')
        if activity_type is None:
            normalized_codename = parameters['notification'].codename.lower()
            activity_type = cls.ACTIVITY_MAP.get(normalized_codename,
                                                 aspy.Create)
        activity_actor = cls.to_activity_object(parameters.get('actor'))
        activity_target = cls.to_activity_object(parameters.get('target'))
        activity = activity_type(
            name=subject, summary=summary,
            object=parameters['activity_object'],
        )
        if activity_actor is not None:
            activity['actor'] = activity_actor
        if activity_target is not None:
            activity['target'] = activity_target
        return str(activity)

    @classmethod
    def preview_message(cls, subject: str, body: str, parameters: dict):
        subject, summary = cls._build(subject, body, parameters)
        return '<div><strong>{}</strong></div>' \
               '<div>{}</div'.format(subject, summary)

    def send_message(self, _from_address: str,
                     to_addresses: List[str], message: str):
        address_set = set(to_addresses)
        # note: we don't support public addressing yet, just strip them
        if PUBLIC_ADDRESS in address_set:
            address_set.remove(PUBLIC_ADDRESS)
        users = id_to_user(address_set)
        json_data = json.loads(message)
        kwargs = {}
        for field in ('actor', 'object', 'target'):
            field_data = json_data.get(field)
            if isinstance(field_data, str):
                kwargs[field + '_id'] = field_data
                kwargs[field + '_json'] = ''
            elif isinstance(field_data, dict):
                field_id = field_data.get('id')
                if field_id is not None:
                    kwargs[field + '_id'] = field_id
                kwargs[field + '_json'] = json.dumps(field_data)

        for field in ('name', 'summary', 'type'):
            if field in json_data:
                kwargs[field] = json_data[field]

        with django.db.transaction.atomic():
            activity = django_vox.models.Activity.objects.create(**kwargs)
            for user in users:
                django_vox.models.InboxItem.objects.create(
                    activity=activity, owner=user)



import importlib.util
import pydoc
import re
import sys

import django.conf

__all__ = ('BACKENDS', 'INBOX_LIMIT', 'THROW_EXCEPTIONS', 'ACTIVITY_REGEX',
           'MARKDOWN_EXTRAS', 'MARKDOWN_LINK_PATTERNS',
           'TWILIO_ACCOUNT_SID', 'TWILIO_AUTH_TOKEN', 'TWILIO_FROM_NUMBER',
           'TWITTER_CONSUMER_KEY', 'TWITTER_CONSUMER_SECRET',
           'TWITTER_TOKEN_KEY', 'TWITTER_TOKEN_SECRET',
           'XMPP_JID', 'XMPP_PASSWORD', 'POSTMARK_TOKEN')


# General settings
BACKENDS = []
INBOX_LIMIT = 500
THROW_EXCEPTIONS = getattr(django.conf.settings, 'DEBUG', False)
ACTIVITY_REGEX = r'activity/(?P<id>[0-9a-z\-]+)/$'

# Markdown settings
URL_PATTERN = re.compile(
    r'((([A-Za-z]{3,9}:(?:\/\/)?)(?:[\-;:&=\+\$,\w]+@)?[A-Za-z0-9\.\-]'
    r'+(:[0-9]+)?|(?:www\.|[\-;:&=\+\$,\w]+@)[A-Za-z0-9\.\-]+)((?:\/[\+'
    r'~%\/\.\w\-_]*)?\??(?:[\-\+=&;%@\.\w_]*)#?(?:[\.\!\/\\\w]*))?)')
MARKDOWN_EXTRAS = ['footnotes', 'link_patterns', 'smarty-pants', 'tables']
MARKDOWN_LINK_PATTERNS = [(URL_PATTERN, r'\1')]

# Twilio
TWILIO_ACCOUNT_SID = None
TWILIO_AUTH_TOKEN = None
TWILIO_FROM_NUMBER = None

# Twitter settings
TWITTER_CONSUMER_KEY = None
TWITTER_CONSUMER_SECRET = None
TWITTER_TOKEN_KEY = None
TWITTER_TOKEN_SECRET = None

# XMPP settings
XMPP_JID = None
XMPP_PASSWORD = None

# Postmark
POSTMARK_TOKEN = None


class _Sneaky:
    def __init__(self, name):
        self.module = sys.modules[name]
        sys.modules[name] = self
        # sphinx would like me to do something like this, but it
        # breaks things right now
        # for name in __all__:
        #     if name != 'BACKENDS':
        #         setattr(self, name, getattr(self.module, name))

    def __getattr__(self, name):
        # call module.__init__ after import introspection is done
        django_setting_name = 'DJANGO_VOX_' + name
        if name in __all__:
            attr = getattr(django.conf.settings, django_setting_name, None)
            if attr is not None:
                return attr
            elif name == 'BACKENDS':
                return _Sneaky.get_backends()
        result = getattr(self.module, name)
        if result is None:
            raise django.conf.ImproperlyConfigured(
                'Please set {} in your settings'.format(django_setting_name))
        return result

    @staticmethod
    def get_backends():
        result = []
        # Automatically set based on libraries available
        default_backends = [
            # activity backend intentionally left out because
            # it requires lot of setup
            'django_vox.backends.html_email.Backend',
            'django_vox.backends.markdown_email.Backend',
            'django_vox.backends.postmark_email.Backend',
            'django_vox.backends.template_email.Backend',
            'django_vox.backends.twilio.Backend',
            'django_vox.backends.twitter.Backend',
            'django_vox.backends.slack.Backend',
            'django_vox.backends.json_webhook.Backend',
            'django_vox.backends.xmpp.Backend',
        ]
        for cls_str in default_backends:
            cls = pydoc.locate(cls_str)
            for dep in cls.DEPENDS:
                if importlib.util.find_spec(dep) is None:
                    continue
            result.append(cls_str)
        return result


_Sneaky(__name__)



from unittest.mock import patch

import django.conf
from django.test import TestCase

import django_vox.backends.twitter


class TestTwitterBackend(TestCase):

    TEXT = 'Here is a text message \n\n for {{ you }}'
    PARAMS = {'you': 'me'}
    SUBJECT = 'IGNORED'
    MESSAGE = 'Here is a text message \n\n for me'
    PREVIEW = 'Here is a text message <br/><br/> for me'

    @classmethod
    def test_build_message(cls):
        backend = django_vox.backends.twitter.Backend
        message = backend.build_message(cls.SUBJECT, cls.TEXT, cls.PARAMS, [])
        assert cls.MESSAGE == message

    @classmethod
    def test_preview_message(cls):
        backend = django_vox.backends.twitter.Backend
        message = backend.preview_message(cls.SUBJECT, cls.TEXT, cls.PARAMS)
        assert cls.PREVIEW == message

    def test_send_message(self):
        backend = django_vox.backends.twitter.Backend
        message = backend.build_message(
            self.SUBJECT, self.TEXT, self.PARAMS, [])
        with patch('twitter.Api'):
            with self.assertRaises(django.conf.ImproperlyConfigured):
                backend()
            with self.settings(
                    DJANGO_VOX_TWITTER_CONSUMER_KEY='consumer key',
                    DJANGO_VOX_TWITTER_CONSUMER_SECRET='consumer secret',
                    DJANGO_VOX_TWITTER_TOKEN_KEY='token key',
                    DJANGO_VOX_TWITTER_TOKEN_SECRET='token secret'):
                instance = backend()
                # post a public message
                instance.send_message('', [], message)
                instance.send_message('', ['foo'], message)
                client = instance.client
                assert len(client.mock_calls) >= 2
                assert client.mock_calls[0][0] == 'PostUpdate'
                assert client.mock_calls[0][1] == (message,)
                assert client.mock_calls[1][0] == 'PostDirectMessage'
                assert client.mock_calls[1][1] == (message,)
                assert client.mock_calls[1][2] == {'screen_name': 'foo'}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models

import django_vox.registry


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('django_vox', '0002_attachments'),
    ]

    operations = [
        migrations.RenameField(
            model_name='notification',
            old_name='target_model',
            new_name='target_type'),
        migrations.RenameField(
            model_name='notification',
            old_name='source_model',
            new_name='actor_type'),
        migrations.RenameField(
            model_name='notification',
            old_name='content_type',
            new_name='object_type'),
        migrations.AlterField(
            model_name='notification',
            name='actor_type',
            field=models.ForeignKey(
                blank=True, null=True,
                limit_choices_to=django_vox.registry.channel_type_limit,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='+', to='contenttypes.ContentType',
                verbose_name='actor model'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='target_type',
            field=models.ForeignKey(
                blank=True, null=True,
                limit_choices_to=django_vox.registry.channel_type_limit,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='+', to='contenttypes.ContentType',
                verbose_name='target model'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='object_type',
            field=models.ForeignKey(
                limit_choices_to=django_vox.registry.channel_type_limit,
                on_delete=django.db.models.deletion.CASCADE,
                to='contenttypes.ContentType', verbose_name='object type'),
        ),
        migrations.AddField(
            model_name='notification',
            name='activity_type',
            field=models.URLField(
                blank=True, default='',
                help_text='Full URL for activity type (i.e. '
                          'https://www.w3.org/ns/activitystreams#Object)',
                verbose_name='activity type'),
        ),
        migrations.AlterField(
            model_name='templateattachment',
            name='template',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='attachments', to='django_vox.Template',
                verbose_name='template'),
        ),
        migrations.CreateModel(
            name='InboxItem',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True, serialize=False,
                    verbose_name='ID')),
                ('timestamp', models.DateTimeField(
                    auto_now_add=True, db_index=True,
                    verbose_name='timestamp')),
                ('json', models.TextField()),
                ('owner_id', models.CharField(
                    verbose_name='owner id', db_index=True, max_length=2048)),
                ('owner_type', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='+', to='contenttypes.ContentType',
                    verbose_name='owner type')),
                ('actor_id', models.CharField(
                    verbose_name='actor id', db_index=True, max_length=2048)),
                ('object_id', models.CharField(
                    verbose_name='object id', db_index=True, max_length=2048)),
                ('type', models.CharField(
                    verbose_name='type', db_index=True, max_length=512)),
            ],
        ),
    ]



"""
Miscellaneous functions and shared classes
"""

from django.apps import apps
from django.conf import settings


def get_current_site():
    """
    Get current site without the request object.
    """
    # Imports are inside the function because its point is to avoid importing
    # the Site models when django.contrib.sites isn't installed.
    if apps.is_installed('django.contrib.sites'):
        from django.contrib.sites.models import Site
        return Site.objects.get_current()
    else:
        return SettingsSite()


class SettingsSite:
    """
    A class that shares the primary interface of Site (from
    django.contrib.sites) but gets its data the config.
    """

    def __init__(self):
        self.domain = self.name = getattr(
            settings, 'SITE_DOMAIN', '127.0.0.1:8000')

    def __str__(self):
        return self.name

    def save(self, force_insert=False, force_update=False):
        raise NotImplementedError('SettingsSite cannot be saved.')

    def delete(self):
        raise NotImplementedError('SettingsSite cannot be deleted.')


def full_iri(absolute_url: str):
    """
    Convert an non-relative url into a fully qualified url

    Uses the SITE_SSL and SITE_DOMAIN settings
    """
    # copied and modified from django.contrib.syndication.view.add_domain
    ssl = getattr(settings, 'SITE_SSL', True)
    protocol = 'https' if ssl else 'http'
    if absolute_url.startswith('//'):
        return '{}:{}'.format(protocol, absolute_url)
    if absolute_url.startswith(('http://', 'https://', 'mailto:')):
        return absolute_url
    domain = get_current_site().domain
    return '{}://{}{}'.format(protocol, domain, absolute_url)


class Contact:
    """A generic contact object

    If you want to return something that looks like this, make sure
    to implement the __hash__ method the same, otherwise filtering
    duplicate contacts won't work
    """

    def __init__(self, name: str, protocol: str, address: str):
        self.name = name
        self.protocol = protocol
        self.address = address

    def __str__(self):
        return '{} <{}:{}>'.format(self.name, self.protocol, self.address)

    def __hash__(self):
        return hash((self.name, self.protocol, self.address))



import abc
import collections
import inspect
import json
import random
import uuid
import warnings
from typing import List, Mapping, TypeVar, cast

import aspy
import dateutil.parser
import django.conf
import django.utils.timezone
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import NOT_PROVIDED, Q
from django.template import Context
from django.utils.translation import ugettext_lazy as _

import django_vox.backends
from django_vox.backends.base import AttachmentData

from . import base, registry, settings

__all__ = ('find_activity_type', 'get_model_from_relation',
           'resolve_parameter', 'make_model_preview', 'PreviewParameters',
           'get_model_variables', 'get_model_attachment_choices',
           'AbstractContactable', 'VoxOptions', 'VoxModelBase', 'VoxModel',
           'VoxNotification', 'VoxNotifications', 'VoxAttach',
           'VoxAttachments', 'NotificationManager', 'Notification',
           'Template', 'TemplateAttachment', 'SiteContactManager',
           'OneTimeNotification',
           'SiteContact', 'SiteContactSetting', 'FailedMessage', 'InboxItem')


def load_aspy_object(json_str):
    data = json.loads(json_str)
    return _load_aspy_object(data)


def _load_aspy_object(data):
    obj = getattr(aspy, data.get('type', 'Object'))()
    for key, value in data.items():
        if key not in ('type', '@context'):
            if aspy.PROPERTIES.get(key) is aspy.datetime_property:
                value = dateutil.parser.parse(value)
            if isinstance(value, dict):
                value = _load_aspy_object(value)
            obj[key] = value
    return obj


def find_activity_type(url):
    if ':' not in url:
        url = aspy.as_uri(url)
    for item in aspy.__dict__.values():
        if inspect.isclass(item):
            if issubclass(item, (aspy.Object, aspy.Link)):
                if item.get_type() == url:
                    return item


def make_activity_object(obj):
    iri = None
    # if obj.__class__ in registry.objects:
    address_func = getattr(obj, 'get_object_address')
    if address_func:
        iri = obj.get_object_address()
    if iri is None and hasattr(obj, 'get_absolute_url'):
        iri = django_vox.base.full_iri(obj.get_absolute_url())
    name = str(obj)
    if iri is None:
        return aspy.Object(name=name)
    return aspy.Object(name=name, id=iri)


def get_model_from_relation(field):
    # code copied from django's admin
    if not hasattr(field, 'get_path_info'):
        raise RuntimeError('Field is not a relation')
    return field.get_path_info()[-1].to_opts.model


def resolve_parameter(key, parameters):
    remainder, _, last = key.rpartition('.')
    parts = remainder.split('.')
    for part in parts:
        try:
            contains = part in parameters
        except TypeError:
            contains = False
        if contains:
            parameters = parameters[part]
        elif hasattr(parameters, part):
            parameters = getattr(parameters, part)
        else:
            return None
    if hasattr(parameters.__class__, '_vox_meta'):
        meta = parameters.__class__._vox_meta
        if last in meta.attachments:
            return meta.attachments[last].get_data(parameters)
    return None


def make_model_preview(object_type):
    obj = object_type.objects.first()
    if obj is not None:
        return obj
    obj = object_type()
    for field in object_type._meta.fields:
        value = None
        if field.default != NOT_PROVIDED:
            value = field.default
        else:
            desc = str(field.description).lower()
            if desc.startswith('string'):
                value = '{{{}}}'.format(field.verbose_name)
            elif desc.startswith('integer'):
                value = random.randint(1, 100)
            else:
                pass
        setattr(obj, field.attname, value)
    return obj


class PreviewParameters:

    def __init__(self, object_type, actor_type, target_type):
        self.target = (make_model_preview(target_type)
                       if target_type else {})
        self.actor = (make_model_preview(actor_type)
                      if actor_type else {})
        self.object = make_model_preview(object_type)
        # for backwards compatibility
        self.source = self.actor
        self.content = self.object

    def __contains__(self, item):
        return item in ('contact', 'target', 'actor', 'object',
                        'content', 'source')

    def __getitem__(self, attr):
        return getattr(self, attr)


def get_model_variables(label, value, cls, ancestors=set()):
    assert issubclass(cls, models.Model)
    sub_ancestors = ancestors.copy()
    sub_ancestors.add(cls)
    attrs = []
    skip_relations = len(ancestors) > 2
    children = []
    for field in cls._meta.fields:
        sub_label = field.verbose_name.title()
        sub_value = '{}.{}'.format(value, field.name)
        if field.is_relation:
            model = get_model_from_relation(field)
            # prevent super long/circular references
            if skip_relations or model in ancestors:
                continue
            children.append(get_model_variables(
                sub_label, sub_value, model, ancestors=sub_ancestors))
        else:
            attrs.append({'label': sub_label, 'value': sub_value})
    return {'label': label, 'value': value, 'attrs': attrs,
            'rels': children}


def get_model_attachment_choices(label, value, cls, ancestors=set()):
    sub_ancestors = ancestors.copy()
    sub_ancestors.add(cls)
    if hasattr(cls, '_vox_meta'):
        fields = cls._vox_meta.attachments
        for field in fields:
            label = ('{}/{}'.format(label, field.label)
                     if label else field.label)
            yield (value + '.' + field.key), label
    if hasattr(cls, '_meta') and len(sub_ancestors) < 3:
        for field in cls._meta.fields:
            if field.is_relation:
                model = get_model_from_relation(field)
                if model not in ancestors:
                    sub_label = field.verbose_name.title()
                    sub_label = ('{}/{}'.format(label, sub_label)
                                 if label else sub_label)
                    sub_value = '{}.{}'.format(value, field.name)
                    yield from get_model_attachment_choices(
                        sub_label, sub_value, field.related_model,
                        ancestors=sub_ancestors)


class AbstractContactable(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_contacts_for_notification(
            self, notification: 'Notification') -> List[base.Contact]:
        ...


class VoxOptions(object):
    """
    Options for Vox extensions
    """

    ALL_OPTIONS = ('notifications', 'attachments')
    # list of notification code names
    notifications = []
    attachments = {}

    def __init__(self, meta):
        """
        Set any options provided, replacing the default values
        """
        if meta is not None:
            for key, value in meta.__dict__.items():
                if key in self.ALL_OPTIONS:
                    setattr(self, key, value)
                elif not key.startswith('_'):  # ignore private parts
                    raise ValueError(
                        'VoxMeta has invalid attribute: {}'.format(key))


class VoxModelBase(models.base.ModelBase):
    """
    Metaclass for Vox extensions.

    Deals with notifications on VoxOptions
    """
    def __new__(mcs, name, bases, attributes):
        new = super(VoxModelBase, mcs).__new__(
            mcs, name, bases, attributes)
        meta = attributes.pop('VoxMeta', None)
        setattr(new, '_vox_meta', VoxOptions(meta))
        return new


VoxModelN = TypeVar('VoxModelN', 'VoxModel', None)


class VoxModel(models.Model, metaclass=VoxModelBase):
    """
    Base class for models that implement notifications
    """

    class Meta:
        abstract = True

    @classmethod
    def get_notification(cls, codename: str) -> 'Notification':
        ct = ContentType.objects.get_for_model(cls)
        return Notification.objects.get(
            codename=codename, object_type=ct)

    def issue_notification(self, codename: str,
                           target: VoxModelN = None,
                           actor: VoxModelN = None):
        notification = self.get_notification(codename)
        notification.issue(self, target, actor)

    def get_activity_object(self, codename, actor, target):
        """Return an aspy.Object object for the activity.

        The parameters actor and target may be None.
        """
        return self.__activity__()

    def get_object_address(self):
        if self.__class__ not in registry.objects:
            raise RuntimeError(
                '{cls} is not a registered object, use '
                'django_vox.registry.object.add({cls}, regex=...)'.format(
                    cls=self.__class__))
        url = registry.objects[self.__class__].reverse(self)
        if url is None:
            return None
        return django_vox.base.full_iri(url)

    def __activity__(self):
        return make_activity_object(self)


class VoxNotifications(list):

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if not isinstance(value, VoxNotification):
                value = VoxNotification(value)
            if not value.params['codename']:
                value.params['codename'] = key
            self.append(value)


class VoxNotification:
    REQUIRED_PARAMS = {'codename', 'description'}
    OPTIONAL_PARAMS = {'actor_type': '', 'target_type': '',
                       'activity_type': 'Create', 'required': False}

    def __init__(self, description, codename='', **kwargs):
        self.params = {
            'codename': codename,
            'description': description,
            'from_code': True,
        }
        for key, default in self.OPTIONAL_PARAMS.items():
            if key in kwargs:
                self.params[key] = kwargs.pop(key)
            else:
                self.params[key] = default
        if kwargs:
            raise ValueError('Unrecognized parameters {}'.format(
                ', '.join(kwargs.keys())))

    def params_equal(self, notification):
        for key in self.params:
            value = getattr(notification, key)
            my_value = self.params[key]
            if key in ('actor_type', 'target_type'):
                if value is None:
                    value = ''
                else:
                    value = '{}.{}'.format(value.app_label, value.model)
            # if key == 'activity_type':
            #     value = value.get_type()
            if value != my_value:
                return False
        return True

    def param_value(self, key):
        value = self.params[key]
        if key in ('actor_type', 'target_type'):
            if value == '':
                return None
            model = apps.get_model(value)
            return ContentType.objects.get_for_model(model)
        if key == 'activity_type':
            if inspect.isclass(value) and issubclass(value, aspy.Object):
                value = value.get_type()
        return value

    def set_params(self, notification):
        for key in self.params:
            setattr(notification, key, self.param_value(key))

    def create(self, object_type):
        new = Notification(object_type=object_type)
        self.set_params(new)
        return new

    @property
    def codename(self):
        return self.params['codename']


# Temporarily here for backwards compatibility
class VoxParam(VoxNotification):

    def __init__(self, codename, description, **kwargs):
        super().__init__(description, codename=codename, **kwargs)


class VoxAttach:
    def __init__(self, attr: str = None,
                 mime_attr: str = '', mime_string: str = '', label=''):
        self.key = ''  # gets set later
        self.attr = attr
        if bool(mime_attr) == bool(mime_string):
            raise RuntimeError(
                'Either mime_attr must be set or mime_string (not both)')
        self.mime_attr = mime_attr
        self.mime_string = mime_string
        self._label = label

    @property
    def label(self):
        return self._label if self._label else self.key

    def get_data(self, model_instance: VoxModel):
        data = getattr(model_instance, self.attr)
        if callable(data):
            data = data()
        # force bytes
        if not isinstance(data, bytes):
            if not isinstance(data, str):
                data = str(data)
            data = data.encode()
        if self.mime_attr:
            mime = getattr(model_instance, self.mime_attr)
            if callable(mime):
                mime = mime()
        else:
            mime = self.mime_string
        return AttachmentData(data, mime)


class VoxAttachments:
    def __init__(self, **kwargs: Mapping[str, VoxAttach]):
        self.items = {}
        for key, value in kwargs.items():
            value.key = key
            if value.attr is None:
                value.attr = key
            self.items[key] = value

    def __iter__(self):
        yield from self.items.values()

    def __contains__(self, item: str):
        return self.items.__contains__(item)

    def __getitem__(self, item: str):
        return self.items.__getitem__(item)


class ChannelContactSet:

    def __init__(self, notification, obj, target, actor):
        # make a dictionary where the keys are protocol, 'recipient key'
        # pairs, the values are address => contactable dictionaries
        channels = notification.get_recipient_channels(obj, target, actor)
        self._set = collections.defaultdict(dict)
        contactable_list = dict((key, channel.contactables())
                                for (key, channel) in channels.items())
        for recipient_key, contactables in contactable_list.items():
            for c_able in contactables:
                for contact in c_able.get_contacts_for_notification(
                        notification):
                    self._set[contact.protocol, recipient_key][
                        contact.address] = c_able

    def get_addresses(self, protocol: str, recipients: List[str]):
        for recipient in recipients:
            if (protocol, recipient) in self._set:
                yield from self._set[protocol, recipient].keys()

    def get_address_items(self, protocol: str, recipients: List[str]):
        for recipient in recipients:
            if (protocol, recipient) in self._set:
                yield from self._set[protocol, recipient].items()


class NotificationManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, app_label, model, codename):
        return self.get(
            codename=codename,
            object_type=ContentType.objects.db_manager(
                self.db).get_by_natural_key(app_label, model),
        )


class Notification(models.Model):
    """
    Base class for all notifications
    """

    codename = models.CharField(_('codename'), max_length=100)
    object_type = models.ForeignKey(
        to=ContentType, on_delete=models.CASCADE,
        limit_choices_to=registry.channel_type_limit,
        verbose_name=_('object type'))
    description = models.TextField(_('description'))
    actor_type = models.ForeignKey(
        to=ContentType, on_delete=models.CASCADE, related_name='+',
        limit_choices_to=registry.channel_type_limit,
        verbose_name=_('actor model'), null=True, blank=True)
    target_type = models.ForeignKey(
        to=ContentType, on_delete=models.CASCADE, related_name='+',
        limit_choices_to=registry.channel_type_limit,
        verbose_name=_('target model'), null=True, blank=True)
    required = models.BooleanField(
        _('required'), default=False,
        help_text=_('If true, triggering the notification will throw an '
                    'error if there is no available template/contact'))
    activity_type = models.URLField(
        _('activity type'), default='', blank=True,
        help_text=_('Full URL for activity type (i.e. '
                    'https://www.w3.org/ns/activitystreams#Object)')
    )
    from_code = models.BooleanField(
        _('from code'), default=False,
        help_text=_('True if the notification is defined in the code and '
                    'automatically managed'))

    objects = NotificationManager()

    def __str__(self):
        return "{} | {} | {}".format(
            self.object_type.app_label, self.object_type, self.codename)

    def natural_key(self):
        return self.object_type.natural_key() + (self.codename,)

    natural_key.dependencies = ['contenttypes.contenttype']

    def get_recipient_models(self):
        recipient_spec = {
            'si': SiteContact,
            're': self.get_target_type(),
            'se': self.get_actor_type(),
            'c': self.get_object_model(),
        }
        return dict((key, value) for (key, value)
                    in recipient_spec.items() if value is not None)

    def get_parameter_models(self):
        spec = {
            'target': self.get_target_type(),
            'actor': self.get_actor_type(),
            'object': self.get_object_model(),
        }
        return dict((key, value) for (key, value)
                    in spec.items() if value is not None)

    def get_recipient_instances(self, obj, target, actor):
        choices = {
            'si': SiteContact(),
            'c': obj,
        }
        if self.target_type and (target is not None):
            choices['re'] = target
        elif self.target_type:
            raise RuntimeError(
                'Model specified "target_type" for notification but '
                'target missing on issue_notification ')
        elif target is not None:
            raise RuntimeError('Target added to issue_notification, '
                               'but is not specified in VoxMeta')
        if self.actor_type and (actor is not None):
            choices['se'] = actor
        elif self.actor_type:
            raise RuntimeError(
                'Model specified "actor_type" for notification but actor '
                'missing on issue_notification ')
        elif actor is not None:
            raise RuntimeError('Actor added to issue_notification, but is '
                               'not specified in VoxMeta')

        return dict((key, model) for (key, model)
                    in choices.items() if model is not None)

    def get_recipient_channels(self, obj, target, actor)\
            -> Mapping[str, registry.Channel]:
        instances = self.get_recipient_instances(obj, target, actor)
        return dict((key, channel)
                    for recip_key, model in instances.items()
                    for key, channel in registry.objects[
                        model.__class__].channels.prefix(
                            recip_key).bind(model).items())

    def get_recipient_choices(self):
        recipient_models = self.get_recipient_models()
        for recipient_key, model in recipient_models.items():
            channel_data = registry.objects[model].channels.prefix(
                recipient_key)
            for key, channel in channel_data.items():
                yield key, channel.name

    def _get_activity_object(self, obj, actor, target):
        if hasattr(obj, 'get_activity_object'):
            return obj.get_activity_object(self.codename, actor, target)
        if hasattr(obj, '__activity__'):
            return obj.__activity__()
        return make_activity_object(obj)

    def issue(self, obj: VoxModel,
              target: VoxModelN = None,
              actor: VoxModelN = None):
        """
        To send a notification to a user, get all the user's active methods.
        Then get the backend for each method and find the relevant template
        to send (and has the said notification). Send that template with
        the parameters with the backend.

        :param obj: model object that the notification is about
        :param target: either a user, or None if no logical target
        :param actor: user who initiated the notification
        :return: None
        """
        # check
        parameters = {
            'content': obj,
            'object': obj,
            'notification': self,
        }
        if target is not None:
            parameters['target'] = target
        if actor is not None:
            parameters['actor'] = actor
            # backwards compatibility
            parameters['source'] = actor

        parameters['activity_object'] = self._get_activity_object(
            obj, actor, target)
        parameters['activity_type'] = self.get_activity_type()
        # load up all the templates so we can see available recipients
        templates = self.template_set.filter(enabled=True)
        contact_set = ChannelContactSet(self, obj, target, actor)

        exceptions = []
        sent = False

        backend_ids = {template.backend for template in templates}
        loaded_backends = dict((bid, django_vox.registry.backends.by_id(bid)())
                               for bid in backend_ids)

        for template in templates:
            backend = loaded_backends[template.backend]
            recipients = template.recipients.split(',')
            # items is a list of address list, contactable pairs
            if template.bulk:
                items = ((contact_set.get_addresses(
                    backend.PROTOCOL, recipients), None),)
            else:
                items = list(((address,), contactable) for
                             (address, contactable) in
                             contact_set.get_address_items(
                                 backend.PROTOCOL, recipients))
            for addresses, contactable in items:
                if contactable is not None:
                    local_params = parameters.copy()
                    local_params['recipient'] = contactable
                else:
                    local_params = parameters

                exception = self.send_message(
                    addresses, local_params, template, backend)
                if exception is None:
                    sent = True
                elif settings.THROW_EXCEPTIONS:
                    exceptions.append(exception)
                else:
                    warnings.warn(RuntimeWarning(exception))

        if exceptions:
            raise exceptions[0]

        if not sent and self.required:
            raise RuntimeError('Notification required, but no message sent')

    @staticmethod
    def send_message(addresses: List[str],
                     parameters: dict, template, backend) -> Exception:
        attachments = []
        if backend.USE_ATTACHMENTS:
            for attachment in template.attachments.all():
                data = resolve_parameter(attachment.key, parameters)
                if data is not None:
                    attachments.append(data)
        message = backend.build_message(
            template.subject, template.content, parameters, attachments)
        from_address = backend.get_from_address(
            template.from_address, parameters)
        # We're catching all exceptions here because some people
        # are bad people and can't subclass properly
        try:
            backend.send_message(from_address, addresses, message)
        except Exception as e:
            FailedMessage.objects.create(
                backend=backend.ID,
                from_address=from_address,
                to_addresses=','.join(addresses),
                message=str(message),
                error=str(e),
            )
            return e
        return None

    def can_issue_custom(self):
        return not (self.actor_type or self.target_type)

    def preview(self, backend_id, message):
        backend = registry.backends.by_id(backend_id)
        params = PreviewParameters(
            self.get_object_model(), self.get_actor_type(),
            self.get_target_type())
        return backend.preview_message('', message, params)

    def get_actor_type(self):
        return (self.actor_type.model_class() if self.actor_type_id
                else None)

    def get_target_type(self):
        return (self.target_type.model_class() if self.target_type_id
                else None)

    def get_object_model(self):
        return (self.object_type.model_class() if self.object_type_id
                else None)

    def get_activity_type(self):
        if self.activity_type == '':
            return aspy.Create
        return find_activity_type(self.activity_type)

    def get_recipient_variables(self):
        recipient_spec = self.get_recipient_models()
        actor_type = self.get_actor_type()
        target_type = self.get_target_type()
        content_model = self.get_object_model()
        mapping = {}
        for target_key, model in recipient_spec.items():
            if model is not None:
                channels = registry.objects[model].channels.prefix(target_key)
                for key, channel in channels.items():
                    label = str(_('Recipient {}')).format(channel.name)
                    mapping[key] = get_model_variables(
                        label, 'recipient', channel.target_class)
        content_name = content_model._meta.verbose_name.title()
        mapping['_static'] = [
            get_model_variables(content_name, 'object', content_model),
        ]
        if actor_type:
            mapping['_static'].append(
                get_model_variables('Actor', 'actor', actor_type))
        if target_type:
            mapping['_static'].append(
                get_model_variables('Target', 'target', target_type))
        return mapping


class _OneTimeNotificationType:
    codename = ''
    object_type_id = None
    description = ''
    actor_type_id = None
    target_type_id = None
    required = True
    activity_type = 'Object'
    from_code = False

    def __str__(self):
        return '<OneTimeNotification>'

    @property
    def object_type(self):
        return None

    @property
    def actor_type(self):
        return None

    @property
    def target_type(self):
        return None

    def natural_key(self):
        return '', '', ''

    def get_actor_type(self):
        return None

    def get_target_type(self):
        return None

    def get_object_model(self):
        return None

    def get_activity_type(self):
        return aspy.Create

    def get_recipient_variables(self):
        return {}

    def send(self, backend_id, contacts, from_address, subject, body):
        backend = django_vox.registry.backends.by_id(backend_id)()
        to_addresses = [c.address for c in contacts
                        if c.protocol == backend.PROTOCOL]

        parameters = {}
        message = backend.build_message(
            subject, body, parameters, attachments=[])
        from_address = backend.get_from_address(from_address, parameters)
        try:
            backend.send_message(from_address, to_addresses, message)
        except Exception as e:
            FailedMessage.objects.create(
                backend=backend.ID,
                from_address=from_address,
                to_addresses=','.join(to_addresses),
                message=str(message),
                error=str(e),
            )
            return e
        return None


OneTimeNotification = _OneTimeNotificationType()


class Template(models.Model):

    class Meta:
        verbose_name = _('template')

    notification = models.ForeignKey(
        to=Notification, on_delete=models.PROTECT,
        verbose_name=_('notification'))
    backend = models.CharField(_('backend'), max_length=100)
    subject = models.CharField(_('subject'), max_length=500, blank=True)
    content = models.TextField(_('content'))
    recipients = models.TextField(
        verbose_name=_('recipients'), default='re', blank=True,
        help_text=_('Who this message actually gets sent to.'))
    from_address = models.CharField(
        _('from address'), max_length=500, blank=True)
    enabled = models.BooleanField(
        _('enabled'), default=True,
        help_text=_('When not active, the template will be ignored'))
    bulk = models.BooleanField(
        _('bulk'), default=True,
        help_text=_('Send the same message to all recipients'))

    objects = models.Manager()

    def render(self, parameters: dict, autoescape=True):
        content = cast(str, self.content)
        template = django_vox.backends.base.template_from_string(content)
        context = Context(parameters, autoescape=autoescape)
        return template.render(context)

    def __str__(self):
        choices = {}
        if self.notification:
            choices = dict(self.notification.get_recipient_choices())
        recipients = ', '.join(str(choices.get(r, r)) for r
                               in self.recipients.split(','))
        backend = registry.backends.by_id(self.backend)
        return '{} for {}'.format(backend.VERBOSE_NAME, recipients)


class TemplateAttachment(models.Model):

    class Meta:
        verbose_name = _('template attachment')

    template = models.ForeignKey(
        to=Template, on_delete=models.CASCADE,
        verbose_name=_('template'), related_name='attachments')
    key = models.CharField(_('key'), max_length=500)


class SiteContactManager(models.Manager, AbstractContactable):
    use_in_migrations = True

    def get_contacts_for_notification(
            self, notification: 'Notification') -> List[base.Contact]:
        wlq = Q(enable_filter='whitelist',
                sitecontactsetting__notification=notification,
                sitecontactsetting__enabled=True)
        blq = Q(enable_filter='blacklist') & ~Q(
                sitecontactsetting__notification=notification,
                sitecontactsetting__enabled=False)
        for sc in SiteContact.objects.filter(blq | wlq).distinct():
            yield base.Contact(sc.name, sc.protocol, sc.address)


# can't make this subclass AbstractContact or fields become unset-able
class SiteContact(VoxModel):

    ENABLE_CHOICES = (
        ('blacklist', _('Blacklist')),
        ('whitelist', _('Whitelist')),
    )

    class Meta:
        verbose_name = _('site contact')
        unique_together = (('address', 'protocol'),)

    name = models.CharField(_('name'), blank=True, max_length=500)
    protocol = models.CharField(_('protocol'), max_length=100)
    address = models.CharField(_('address'), max_length=500, blank=True)
    enable_filter = models.CharField(choices=ENABLE_CHOICES, max_length=10,
                                     default='blacklist')

    objects = SiteContactManager()

    def __str__(self):
        return self.name

    @staticmethod
    def all_contacts(_obj):
        yield SiteContact.objects


class SiteContactSetting(models.Model):

    class Meta:
        verbose_name = _('site contact setting')
        unique_together = (('site_contact', 'notification'),)

    site_contact = models.ForeignKey(SiteContact, on_delete=models.CASCADE)
    notification = models.ForeignKey(Notification, on_delete=models.CASCADE)
    enabled = models.BooleanField(_('enabled'))

    objects = models.Manager()


class FailedMessage(models.Model):

    class Meta:
        verbose_name = _('failed message')

    backend = models.CharField(_('backend'), max_length=100)
    from_address = models.CharField(_('from address'),
                                    max_length=500, blank=True)
    to_addresses = models.TextField(_('to addresses'))
    message = models.TextField(_('message'))
    error = models.TextField(_('error'))
    created_at = models.DateTimeField(_('created at'), auto_now_add=True)

    objects = models.Manager()

    def __str__(self):
        return '{} @ {}'.format(self.to_addresses, self.created_at)

    def resend(self):
        backend = django_vox.registry.backends.by_id(self.backend)()
        to_addresses = list(self.to_addresses.split(','))
        backend.send_message(self.from_address, to_addresses, self.message)
        self.delete()


class Activity(models.Model):
    id = models.UUIDField(
        _('ID'), primary_key=True, serialize=True, default=uuid.uuid4)
    name = models.CharField(
        _('name'), max_length=512, blank=True)
    summary = models.TextField(blank=True)
    type = models.CharField(
        _('type'), max_length=512, db_index=True)
    # the json fields should be able to be referenced and
    # repopulated by the id fields
    actor_id = models.CharField(
        _('actor id'), max_length=2048, db_index=True)
    actor_json = models.TextField(blank=True)
    target_id = models.CharField(
        _('target id'), max_length=2048, db_index=True)
    target_json = models.TextField(blank=True)
    object_id = models.CharField(
        _('object id'), max_length=2048, db_index=True)
    object_json = models.TextField(blank=True)
    # and a timestamp, maybe useful
    timestamp = models.DateTimeField(
        _('timestamp'), db_index=True, auto_now_add=True)

    def get_full_url(self):
        return django_vox.base.full_iri(
            registry.objects[Activity].reverse(self))

    def __activity__(self):
        # this is a bit hackish because we're using dicts and not aspy
        # objects
        aspy_class = getattr(aspy, self.type, aspy.Object)
        obj = aspy_class(id=self.get_full_url(), published=self.timestamp)
        for field in ('actor', 'object', 'target'):
            value = getattr(self, field + '_json')
            if value:
                obj[field] = load_aspy_object(value)
        for field in 'summary', 'name':
            value = getattr(self, field)
            if value:
                obj[field] = value
        return obj

    def activity_read(self, _activity, user):
        inbox_item = InboxItem.objects.get(owner=user, activity=self)
        inbox_item.read_at = django.utils.timezone.now()
        inbox_item.save()


class InboxItem(models.Model):

    class Meta:
        unique_together = ('owner', 'activity')

    owner = models.ForeignKey(
        to=django.conf.settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE, related_name='+',
        verbose_name=_('owner'))
    activity = models.ForeignKey(
        to=Activity, on_delete=models.CASCADE, related_name='+',
        verbose_name=_('activity'))
    read_at = models.DateTimeField(
        _('date read'), db_index=True, null=True, blank=True)


registry.objects.add(SiteContact, regex=None)
registry.objects[SiteContact].channels.add(
    '', '', SiteContact, SiteContact.all_contacts)
registry.objects.add(Activity, regex=settings.ACTIVITY_REGEX)



from datetime import datetime
from unittest.mock import patch

from django.test import TestCase

from django_vox.backends.json_webhook import Backend

from tests.mock import MockJsonResponse


def mocked_requests_post(url, _data=None, json=None, **_kwargs):
    if url == 'https://webhook.example':
        return MockJsonResponse('', 200)
    return MockJsonResponse(None, 404)


class TestJsonWebhookBackend(TestCase):

    TEXT = 'birthday : {{ birthday | date:"r" }}\n' \
           'empty: {{ null_value | default:"blank" }}\n' \
           'html: <i>{{ html }}</i>\n' \
           '\n'
    PARAMS = {
        'birthday': datetime(1984, 12, 11),
        'html': '<b>BOO!</b>',
        'null_value': None}
    MESSAGE = '<html>\n' \
              '<dl>\n' \
              '<dt>birthday</dt><dd>Tue, 11 Dec 1984 00:00:00 -0600</dd>\n' \
              '<dt>empty</dt><dd>blank</dd>\n' \
              '<dt>html</dt><dd><i>&lt;b&gt;BOO!&lt;/b&gt;</i></dd>\n' \
              '</dl>\n' \
              '</html>'
    PREVIEW = MESSAGE

    @classmethod
    def test_build_message(cls):
        message = Backend.build_message('', cls.TEXT, cls.PARAMS, [])
        assert cls.MESSAGE == message

    @classmethod
    def test_preview_message(cls):
        message = Backend.preview_message('', cls.TEXT, cls.PARAMS)
        assert cls.PREVIEW == message

    def test_send_message(self):
        message = Backend.build_message('', self.TEXT, self.PARAMS, [])
        bad_address = 'https://not.example'
        address = 'https://webhook.example'
        with patch('requests.post', side_effect=mocked_requests_post):
            instance = Backend()
            with self.assertRaises(RuntimeError):
                instance.send_message('', [bad_address], message)
            instance.send_message('', [address], message)
            import requests
            check_model = requests.post.mock_calls[0][2]['json']
            assert check_model['birthday'] == \
                'Tue, 11 Dec 1984 00:00:00 -0600'
            assert check_model['empty'] == 'blank'
            # not sure if this is actually what we want, but it's what
            # currently happens
            assert check_model['html'] is None

    def test_from_address(self):
        # we don't use from addresses, so this should be blank
        assert '' == Backend.get_from_address('foo', {})



from django.utils.translation import ugettext_lazy as _

from django_vox import settings

from . import base

__all__ = ('Backend',)


class Backend(base.Backend):

    ID = 'twilio'
    PROTOCOL = 'sms'
    USE_FROM_ADDRESS = True
    ESCAPE_HTML = False
    EDITOR_TYPE = 'basic'
    VERBOSE_NAME = _('Twilio')
    DEPENDS = ('twilio',)

    def __init__(self):
        from twilio.rest import Client
        self.client = Client(
            username=settings.TWILIO_ACCOUNT_SID,
            password=settings.TWILIO_AUTH_TOKEN)

    def send_message(self, from_address, to_addresses, message):
        for address in to_addresses:
            self.client.messages.create(
                to=address, from_=from_address, body=message)

    @classmethod
    def get_default_from_address(cls):
        return settings.TWILIO_FROM_NUMBER



from unittest.mock import patch

import django.conf
from django.test import TestCase

from django_vox.backends.postmark_email import Backend

from tests.mock import MockJsonResponse


def mocked_requests_post(url, _data=None, json=None, **_kwargs):
    if url == Backend.ENDPOINT:
        if not (json.get('TemplateAlias') or json.get('TemplateId')):
            return MockJsonResponse(
                {'ErrorCode': "403", 'Message': 'details'}, 422)
        else:
            return MockJsonResponse({
                "To": "george@example.test",
                "SubmittedAt": "2014-02-17T07:25:01.4178645-05:00",
                "MessageID": "0a129aee-e1cd-480d-b08d-4f48548ff48d",
                "ErrorCode": 0, "Message": "OK"}, 200)
    return MockJsonResponse(None, 404)


class TestPostmarkBackend(TestCase):

    TEXT = 'line 1 : {{ line_1 }}\n' \
           'line 2 : {{ line_2 }}\n' \
           'line 3 : {{ line_3 }}\n' \
           'line 4 : {{ line_4 }}\n' \
           '\n\n' \
           'c\'est vide'
    PARAMS = {
        'line_1': 'poisson un', 'line_2': 'poisson deux',
        'line_3': 'poisson rouge', 'line_4': 'poisson bleu'}
    SUBJECT = 'SUBJECT'
    MESSAGE = '<html>\n' \
              '<h1>SUBJECT</h1>\n' \
              '<dl>\n' \
              '<dt>line 1</dt><dd>poisson un</dd>\n' \
              '<dt>line 2</dt><dd>poisson deux</dd>\n' \
              '<dt>line 3</dt><dd>poisson rouge</dd>\n' \
              '<dt>line 4</dt><dd>poisson bleu</dd>\n' \
              '<dt>c\'est vide</dt><dd></dd>\n' \
              '</dl>\n' \
              '</html>'
    PREVIEW = MESSAGE

    @classmethod
    def test_build_message(cls):
        backend = Backend()
        message = backend.build_message(
            cls.SUBJECT, cls.TEXT, cls.PARAMS, [])
        assert cls.MESSAGE == message

    @classmethod
    def test_preview_message(cls):
        backend = Backend()
        message = backend.preview_message(cls.SUBJECT, cls.TEXT, cls.PARAMS)
        assert cls.PREVIEW == message

    def test_send_message(self):
        backend = Backend
        bad_message = backend.build_message(
            '', self.TEXT, self.PARAMS, [])
        message = backend.build_message(
            self.SUBJECT, self.TEXT, self.PARAMS, [])
        from_address = backend.get_default_from_address()
        to_addresses = ['george@example.test']
        with patch('requests.post', side_effect=mocked_requests_post):
            instance = backend()
            with self.assertRaises(django.conf.ImproperlyConfigured):
                instance.send_message(from_address, to_addresses, message)
            with self.settings(DJANGO_VOX_POSTMARK_TOKEN='token'):
                with self.assertRaises(RuntimeError):
                    instance.send_message(from_address, to_addresses,
                                          bad_message)
                instance.send_message(from_address, to_addresses, message)
                import requests
                call_json = requests.post.mock_calls[1][2]['json']
                assert 'admin@example.test' == call_json['From']
                assert 'george@example.test' == call_json['To']
                check_model = call_json['TemplateModel']
                assert check_model['line 1'] == 'poisson un'
                assert check_model['line 2'] == 'poisson deux'
                assert "c'est vide" in check_model



from django.test.utils import override_settings

import django_vox.settings


def test_backends():
    with override_settings(DJANGO_VOX_BACKENDS=None):
        backends = django_vox.settings.BACKENDS
        # we should have all the backend dependencies installed for
        # testing purposes
        expected = [
            'django_vox.backends.html_email.Backend',
            'django_vox.backends.markdown_email.Backend',
            'django_vox.backends.postmark_email.Backend',
            'django_vox.backends.template_email.Backend',
            'django_vox.backends.twilio.Backend',
            'django_vox.backends.twitter.Backend',
            'django_vox.backends.slack.Backend',
            'django_vox.backends.json_webhook.Backend',
            'django_vox.backends.xmpp.Backend',
        ]
        assert expected == backends



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-04-27 18:10
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models

import django_vox.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='FailedMessage',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('backend', models.CharField(
                    max_length=100, verbose_name='backend')),
                ('contact_name', models.CharField(
                    max_length=500, verbose_name='contact name')),
                ('address', models.CharField(
                    max_length=500, verbose_name='address')),
                ('message', models.TextField(verbose_name='message')),
                ('error', models.TextField(verbose_name='error')),
                ('created_at', models.DateTimeField(
                    auto_now_add=True, verbose_name='created at')),
            ],
            options={
                'verbose_name': 'failed message',
            },
        ),
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('codename', models.CharField(
                    max_length=100, verbose_name='codename')),
                ('description', models.TextField(
                    verbose_name='description')),
                ('required', models.BooleanField(
                    default=False,
                    help_text='If true, triggering the notification '
                              'will throw an error if there is no '
                              'available template/contact',
                    verbose_name='required')),
                ('from_code', models.BooleanField(
                    default=False,
                    help_text='True if the notification is defined in '
                              'the code and automatically managed',
                    verbose_name='from code')),
                ('content_type', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='contenttypes.ContentType',
                    verbose_name='content type')),
                ('source_model', models.ForeignKey(
                    blank=True, null=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='+', to='contenttypes.ContentType',
                    verbose_name='source model')),
                ('target_model', models.ForeignKey(
                    blank=True, null=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='+', to='contenttypes.ContentType',
                    verbose_name='target model')),
            ],
            managers=[
                ('objects', django_vox.models.NotificationManager()),
            ],
        ),
        migrations.CreateModel(
            name='SiteContact',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('name', models.CharField(
                    blank=True, max_length=500, verbose_name='name')),
                ('protocol', models.CharField(
                    max_length=100, verbose_name='protocol')),
                ('address', models.CharField(
                    max_length=500, verbose_name='address')),
                ('enable_filter', models.CharField(
                    choices=[('blacklist', 'Blacklist'),
                             ('whitelist', 'Whitelist')],
                    default='blacklist', max_length=10)),
            ],
            options={
                'verbose_name': 'site contact',
            },
            managers=[
                ('objects', django_vox.models.SiteContactManager()),
            ],
        ),
        migrations.CreateModel(
            name='SiteContactSetting',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('enabled', models.BooleanField(verbose_name='enabled')),
                ('notification', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='django_vox.Notification')),
                ('site_contact', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='django_vox.SiteContact')),
            ],
            options={
                'verbose_name': 'site contact setting',
            },
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('backend', models.CharField(
                    max_length=100, verbose_name='backend')),
                ('subject', models.CharField(
                    blank=True, max_length=500, verbose_name='subject')),
                ('content', models.TextField(verbose_name='content')),
                ('recipient', models.CharField(
                    default='re',
                    help_text='Who this message actually gets sent to.',
                    max_length=103, verbose_name='recipient')),
                ('enabled', models.BooleanField(
                    default=True,
                    help_text='When not active, the template will be ignored',
                    verbose_name='enabled')),
                ('notification', models.ForeignKey(
                    on_delete=django.db.models.deletion.PROTECT,
                    to='django_vox.Notification',
                    verbose_name='notification')),
            ],
            options={
                'verbose_name': 'template',
            },
        ),
        migrations.AlterUniqueTogether(
            name='sitecontact',
            unique_together={('address', 'protocol')},
        ),
        migrations.AlterUniqueTogether(
            name='sitecontactsetting',
            unique_together={('site_contact', 'notification')},
        ),
    ]



import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_vox', '0004_activity_fields'),
    ]

    operations = [
        migrations.DeleteModel('InboxItem'),
        migrations.CreateModel(
            name='Activity',
            fields=[
                ('id', models.UUIDField(
                    default=uuid.uuid4,
                    primary_key=True, serialize=True, verbose_name='ID')),
                ('name', models.CharField(
                    blank=True, max_length=512, verbose_name='name')),
                ('summary', models.TextField(blank=True)),
                ('type', models.CharField(
                    db_index=True, max_length=512, verbose_name='type')),
                ('actor_id', models.CharField(
                    db_index=True, max_length=2048, verbose_name='actor id')),
                ('actor_json', models.TextField(blank=True)),
                ('target_id', models.CharField(
                    db_index=True, max_length=2048, verbose_name='target id')),
                ('target_json', models.TextField(blank=True)),
                ('object_id', models.CharField(
                    db_index=True, max_length=2048, verbose_name='object id')),
                ('object_json', models.TextField(blank=True)),
                ('timestamp', models.DateTimeField(
                    auto_now_add=True, db_index=True,
                    verbose_name='timestamp')),
            ],
        ),
        migrations.CreateModel(
            name='InboxItem',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True, serialize=False,
                    verbose_name='ID')),
                ('activity', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='+', to='django_vox.Activity',
                    verbose_name='activity')),
                ('owner', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='+', to=settings.AUTH_USER_MODEL,
                    verbose_name='owner')),
                ('read', models.BooleanField(default=False)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='inboxitem',
            unique_together=set([('owner', 'activity')]),
        ),
    ]



"""
Test a few features in the base backend that wouldn't normally get tested
"""
from django.test import TestCase
from django_vox.backends import base

messages = []


class Backend(base.Backend):

    ESCAPE_HTML = True


class TestTestBackend(TestCase):

    TEXT = 'Here is a text message <br/><br/> for {{ you }}'
    PARAMS = {'you': 'me'}
    MESSAGE = 'Here is a text message <br/><br/> for me'
    PREVIEW = 'Here is a text message <br/><br/> for me'

    @classmethod
    def test_build_message(cls):
        message = Backend.build_message('', cls.TEXT, cls.PARAMS, [])
        assert cls.MESSAGE == message

    @classmethod
    def test_preview_message(cls):
        message = Backend.preview_message('', cls.TEXT, cls.PARAMS)
        assert cls.PREVIEW == message

    def test_from_address(self):
        assert '' == Backend.get_default_from_address()



from typing import List

import django.conf
import django.core.mail.backends.base
import django.core.mail.backends.smtp
import django.template
import django.utils.html
from django.template import Context

__ALL__ = ('Backend', 'template_from_string')


def html_format(text: str):
    escaped = django.utils.html.escape(text)
    return escaped.replace('\r\n', '<br/>').replace('\n', '<br/>')


class AttachmentData:
    def __init__(self, data: bytes, mime: str):
        self.data = data
        self.mime = mime


class Backend:

    USE_SUBJECT = False
    USE_ATTACHMENTS = False
    USE_FROM_ADDRESS = False
    ESCAPE_HTML = True
    DEPENDS = ()
    EDITOR_TYPE = 'basic'

    def __init__(self):
        """
        Import any modules or initialize any resources you need for sending.

        This is only used when actually sending messages and not used for
        building or previewing them. The lifetime of a backend object isn't
        all that long right now.
        """
        pass

    @classmethod
    def build_message(cls, subject: str, body: str, parameters: dict,
                      attachments: List[AttachmentData]):
        template = template_from_string(body)
        context = Context(parameters, autoescape=cls.ESCAPE_HTML)
        return template.render(context)

    @classmethod
    def preview_message(cls, subject: str, body: str, parameters: dict):
        message = cls.build_message(subject, body, parameters, [])
        if not cls.ESCAPE_HTML:
            message = html_format(message)
        return message

    @classmethod
    def get_from_address(cls, provided_address, parameters: dict):
        if not cls.USE_FROM_ADDRESS:
            return ''
        if not provided_address:
            return cls.get_default_from_address()
        context = Context(parameters, autoescape=False)
        return template_from_string(provided_address).render(context)

    @classmethod
    def get_default_from_address(cls) -> str:
        """
        Gets the default from address

        You only need to implement this is you have
        USE_FROM_ADDRESS = True. This base method should normally never
        get called (unless an empty from address is normal for your backend).
        """
        return ''

    # instance methods
    def send_message(self, from_address: str,
                     to_addresses: List[str], message: str):
        raise NotImplementedError


def template_from_string(text: str, using=None) -> \
        django.template.base.Template:
    """
    Convert a string into a template object,
    using a given template engine or using the default backends
    from settings.TEMPLATES if no engine was specified.
    """
    # This function is based on django.template.loader.get_template,
    # but uses Engine.from_string instead of Engine.get_template.
    engines = django.template.engines
    engine_list = engines.all() if using is None else [engines[using]]
    exception = None
    for engine in engine_list:
        try:
            return engine.from_string(text).template
        except django.template.exceptions.TemplateSyntaxError as e:
            exception = e
    raise exception



from unittest.mock import patch

import django.conf
from django.test import TestCase

from django_vox.backends.twilio import Backend


class TestTwilioBackend(TestCase):

    TEXT = 'Here is a text message \n\n for {{ you }}'
    PARAMS = {'you': 'me'}
    SUBJECT = 'IGNORED'
    MESSAGE = 'Here is a text message \n\n for me'
    PREVIEW = 'Here is a text message <br/><br/> for me'

    @classmethod
    def test_build_message(cls):
        message = Backend.build_message(cls.SUBJECT, cls.TEXT, cls.PARAMS, [])
        assert cls.MESSAGE == message

    @classmethod
    def test_preview_message(cls):
        message = Backend.preview_message(cls.SUBJECT, cls.TEXT, cls.PARAMS)
        assert cls.PREVIEW == message

    def test_send_message(self):
        message = Backend.build_message(
            self.SUBJECT, self.TEXT, self.PARAMS, [])
        with patch('twilio.rest.Client'):
            with self.assertRaises(django.conf.ImproperlyConfigured):
                Backend()
            with self.settings(DJANGO_VOX_TWILIO_ACCOUNT_SID='abc',
                               DJANGO_VOX_TWILIO_AUTH_TOKEN='secret',
                               DJANGO_VOX_TWILIO_FROM_NUMBER='+321'):
                instance = Backend()
                from_address = Backend.get_from_address('', {})
                instance.send_message(from_address, ['+123'], message)
                import twilio.rest
                client = twilio.rest.Client
                assert len(client.mock_calls) > 1
                assert client.mock_calls[0][0] == ''  # class instantiation
                fname, args, kwargs = client.mock_calls[1]
                assert fname == '().messages.create'
                assert args == ()
                assert len(kwargs) == 3
                assert '+123' == kwargs['to']
                assert '+321' == kwargs['from_']
                assert self.MESSAGE == kwargs['body']



from unittest.mock import patch

from django.test import TestCase

from django_vox.backends.slack import Backend

from tests.mock import MockTextResponse


def mocked_requests_post(url, _data=None, json=None, **_kwargs):
    if url.startswith('https://hooks.slack.com/services/'):
        return MockTextResponse('ok', 200)
    return MockTextResponse('', 404)


class TestSlackBackend(TestCase):

    TEXT = 'Here is a text message \n\n for {{ you }}'
    PARAMS = {'you': 'me'}
    MESSAGE = 'Here is a text message \n\n for me'
    PREVIEW = 'Here is a text message <br/><br/> for me'

    def test_send_message(self):
        message = Backend.build_message('', self.TEXT, self.PARAMS, [])
        with patch('requests.post', side_effect=mocked_requests_post):
            webhook = 'https://hooks.slack.com/services/T07US72AH/' \
                      'B91UK823W/bzJGakQaBMDqM7LFC8apWbdB'
            bad_address = 'https://hooks.slac.com/services/T07US72AH/' \
                          'B91UK823W/bzJGakQaBMDqM7LFC8apWbdB'
            instance = Backend()
            import requests
            with self.assertRaises(requests.HTTPError):
                instance.send_message('', [bad_address], message)
            instance.send_message('', [webhook], message)

            assert 2 == len(requests.post.mock_calls)
            call_args = requests.post.mock_calls[1][2]

            assert '{"text": "Here is a text message \\n\\n for me"}' == \
                   call_args['data']
            assert 'application/json' == call_args['headers']['Content-Type']



from django.utils.translation import ugettext_lazy as _

from django_vox import settings

from . import base

__all__ = ('Backend',)


class Backend(base.Backend):

    ID = 'twitter'
    PROTOCOL = 'twitter'
    EDITOR_TYPE = 'basic'
    ESCAPE_HTML = False
    VERBOSE_NAME = _('Twitter')
    DEPENDS = ('twitter',)

    def __init__(self):
        import twitter
        self.client = twitter.Api(
            consumer_key=settings.TWITTER_CONSUMER_KEY,
            consumer_secret=settings.TWITTER_CONSUMER_SECRET,
            access_token_key=settings.TWITTER_TOKEN_KEY,
            access_token_secret=settings.TWITTER_TOKEN_SECRET)

    def send_message(self, _from_address, to_addresses, message):
        if not to_addresses:
            self.client.PostUpdate(message)
        else:
            for address in to_addresses:
                self.client.PostDirectMessage(message, screen_name=address)



from django.apps import apps
from django.contrib.contenttypes.models import ContentType

import django_vox.models


class BackgroundVoxModel(django_vox.models.VoxModel):
    """
    A vox model that sends its calls though a background task
    """

    class Meta:
        abstract = True

    def issue_notification(self, codename: str,
                           target: django_vox.models.VoxModelN = None,
                           actor: django_vox.models.VoxModelN = None):
        kwargs = {}
        self_cls_str = str(self.__class__._meta)
        if target is not None:
            kwargs['target_cls_str'] = str(target.__class__._meta)
            kwargs['target_id'] = target.pk
        if actor is not None:
            kwargs['actor_cls_str'] = str(actor.__class__._meta)
            kwargs['actor_id'] = actor.pk
        issue_notification(codename, self_cls_str, self.pk, **kwargs)


def issue_notification(
        codename: str,
        object_cls_str: str,
        object_id: int,
        target_cls_str: str = '',
        target_id: int = 0,
        actor_cls_str: str = '',
        actor_id: int = 0):
    object_model = apps.get_model(object_cls_str)
    obj = object_model.objects.get(pk=object_id)
    object_ct = ContentType.objects.get_for_model(obj)

    target = None
    if target_cls_str != '':
        model = apps.get_model(target_cls_str)
        target = model.objects.get(pk=target_id)
    actor = None
    if actor_cls_str != '':
        model = apps.get_model(actor_cls_str)
        actor = model.objects.get(pk=actor_id)

    notification = django_vox.models.Notification.objects.get(
        codename=codename, object_type=object_ct)
    notification.issue(obj, target, actor)


try:
    from background_task import background
    issue_notification = background(queue='django-vox')(issue_notification)
except ImportError:
    pass
except RuntimeError:
    pass



from django_vox import base


def test_domain():
    site = base.get_current_site()
    assert '127.0.0.1:8000' == site.domain
    assert '127.0.0.1:8000' == str(site)


def test_full_iri():
    data = (
        ('http://127.0.0.1:8000/foo/bar', '/foo/bar'),
        ('http://foo/bar', '//foo/bar'),
        ('http://bar', 'http://bar'),
    )
    for expected, url in data:
        assert expected == base.full_iri(url)


def test_contact():
    contact = base.Contact('John Doe', 'email', 'john@doe.name')
    assert 'John Doe <email:john@doe.name>' == str(contact)
    assert hash(('John Doe', 'email', 'john@doe.name')) == hash(contact)



import django.contrib.messages
import django.forms.utils
import django.http
from django import forms
from django.conf.urls import url
from django.contrib import admin
from django.contrib.admin.helpers import Fieldset, ACTION_CHECKBOX_NAME
from django.contrib.admin.options import IS_POPUP_VAR
from django.contrib.admin.utils import unquote
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.core.exceptions import PermissionDenied, ValidationError
from django.db.models import Q
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.html import escape
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from . import models, registry


class SelectWithSubjectData(forms.Select):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_subjects = {}
        self.use_attachments = {}
        self.use_from_address = {}

    def create_option(self, name, value, label, selected,
                      index, subindex=None, attrs=None):
        index = str(index) if subindex is None else "%s_%s" % (index, subindex)
        option_attrs = {
            'data-editor': self.editor_types.get(value),
        }
        data_use_fields = {'data-subject': self.use_subjects,
                           'data-from_address': self.use_from_address,
                           'data-attachment': self.use_attachments}
        for key, item in data_use_fields.items():
            option_attrs[key] = ('true' if item.get(value) else 'false')
        if selected:
            option_attrs.update(self.checked_attribute)
        return {
            'name': name,
            'value': value,
            'label': label,
            'selected': selected,
            'index': index,
            'attrs': option_attrs,
            'type': self.input_type,
            'template_name': self.option_template_name,
        }


class BackendChoiceField(forms.ChoiceField):
    widget = SelectWithSubjectData

    def __init__(self, choices=(), *args, **kwargs):
        backs = list(choices)
        choice_pairs = [(back.ID, back.VERBOSE_NAME) for back in backs]
        use_subjects = dict([(back.ID, back.USE_SUBJECT) for back in backs])
        use_from_address = dict([(back.ID, back.USE_FROM_ADDRESS)
                                 for back in backs])
        use_attachments = dict([(back.ID, back.USE_ATTACHMENTS)
                                for back in backs])
        editor_types = dict([(back.ID, back.EDITOR_TYPE) for back in backs])
        super().__init__(choices=choice_pairs, *args, **kwargs)
        self.widget.use_subjects = use_subjects
        self.widget.use_attachments = use_attachments
        self.widget.use_from_address = use_from_address
        self.widget.editor_types = editor_types

    def set_backend_choices(self, backends):
        choice_pairs = [(back.ID, back.VERBOSE_NAME) for back in backends]
        use_subjects = dict([(back.ID, back.USE_SUBJECT) for back in backends])
        use_from_address = dict([(back.ID, back.USE_FROM_ADDRESS)
                                 for back in backends])
        use_attachments = dict([(back.ID, back.USE_ATTACHMENTS)
                                for back in backends])
        editor_types = dict([(back.ID, back.EDITOR_TYPE) for back in backends])
        self.choices = choice_pairs
        self.widget.use_subjects = use_subjects
        self.widget.use_attachments = use_attachments
        self.widget.use_from_address = use_from_address
        self.widget.editor_types = editor_types


class TemplateInlineFormSet(forms.BaseInlineFormSet):
    def get_form_kwargs(self, index):
        kwargs = super().get_form_kwargs(index)
        kwargs['notification'] = self.instance
        return kwargs


class TemplateForm(forms.ModelForm):
    backend = BackendChoiceField(choices=registry.backends.all())
    recipients = forms.MultipleChoiceField(
        widget=forms.CheckboxSelectMultiple, required=False)
    attachments = forms.MultipleChoiceField(
        required=False,
        widget=FilteredSelectMultiple(
            verbose_name=_('Attachments'), is_stacked=False))

    class Meta:
        model = models.Template
        fields = ['notification', 'backend', 'recipients', 'subject',
                  'content', 'attachments', 'from_address', 'bulk', 'enabled']

    def __init__(self, notification=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if notification is not None:
            self.fields['recipients'].choices = \
                notification.get_recipient_choices()
            self.fields['attachments'].choices = \
                self.get_attachment_choices(notification)
        self.initial['attachments'] = [
            a.key for a in self.instance.attachments.all()]
        self.initial['recipients'] = self.instance.recipients.split(',')

    def save(self, commit=True):
        # data will be a set of the added items
        self.instance.recipients = ','.join(
            set(self.cleaned_data['recipients']))
        instance = super().save(commit=commit)
        if commit:
            attach_data = set(self.cleaned_data['attachments'])
            # we need to sync it with the saved items
            query = ~Q(key__in=attach_data) & Q(template=instance)
            models.TemplateAttachment.objects.filter(query).delete()
            added_attachments = []
            for attachment in instance.attachments.all():
                if attachment.key in attach_data:
                    attach_data.remove(attachment.key)
            for key in attach_data:
                added_attachments.append(models.TemplateAttachment(
                    template=instance, key=key))
            if added_attachments:
                models.TemplateAttachment.objects.bulk_create(
                    added_attachments)
        return instance

    def clean(self):
        data = self.cleaned_data
        notification = data['notification']
        try:
            notification.preview(data['backend'], data['content'])
        except Exception as exc:
            raise ValidationError(str(exc))

    @staticmethod
    def get_attachment_choices(notification):
        param_models = notification.get_parameter_models()
        for model_key, model in param_models.items():
            label = '' if model_key == 'object' else model_key.title()
            yield from models.get_model_attachment_choices(
                label, model_key, model)


class SiteContactForm(forms.ModelForm):
    protocol = forms.ChoiceField(choices=registry.get_protocol_choices())

    class Meta:
        model = models.SiteContact
        fields = ['name', 'protocol', 'address', 'enable_filter']


class NotificationForm(forms.ModelForm):

    class Meta:
        model = models.Notification
        fields = ['codename', 'object_type', 'description', 'target_type']


class NotificationIssueForm(forms.Form):

    objects = forms.ModelMultipleChoiceField(
        queryset=None, label=_('Objects'),
        widget=FilteredSelectMultiple(
            verbose_name=_('Objects'), is_stacked=False),
        help_text=_('A separate notification will be issued for each object.')
    )

    def __init__(self, notification, *args, **kwargs):
        self.notification = notification
        model_cls = notification.object_type.model_class()
        self.declared_fields['objects'].queryset = model_cls.objects
        super().__init__(*args, **kwargs)

    def save(self, commit=True):
        if commit:
            objects = self.cleaned_data['objects']
            for obj in objects:
                self.notification.issue(obj)


class TemplateInline(admin.StackedInline):
    model = models.Template
    form = TemplateForm
    formset = TemplateInlineFormSet
    min_num = 0
    extra = 0


class NotificationAdmin(admin.ModelAdmin):
    change_form_template = 'django_vox/change_form.html'
    list_display = ('__str__', 'description', 'required', 'template_count')
    list_filter = ('object_type',)
    inlines = [TemplateInline]
    form = NotificationForm
    issue_form = NotificationIssueForm
    issue_template = 'django_vox/issue.html'

    fields = ['codename', 'object_type', 'description']

    class Media:
        css = {
            'all': ('django_vox/markitup/images.css',
                    'django_vox/markitup/style.css'),
        }
        js = (
            'admin/js/jquery.init.js',
            'django_vox/markitup/jquery.markitup.js',
            'django_vox/notification_fields.js',
        )

    # only show inlines on change forms
    def get_inline_instances(self, request, obj=None):
        return obj and super().get_inline_instances(request, obj) or []

    def get_urls(self):
        return [
            url(
                r'^preview/(?P<backend_id>.+)/$',
                self.admin_site.admin_view(self.preview),
                name='django_vox_preview'),
            url(
                r'^(?P<notification_id>\w+)/preview/(?P<backend_id>.+)/$',
                self.admin_site.admin_view(self.notification_preview),
                name='django_vox_notification_preview'),
            url(
                r'^(?P<notification_id>\w+)/variables/$',
                self.admin_site.admin_view(self.variables),
                name='django_vox_variables'),
            url(
                r'^(?P<notification_id>\w+)/issue/$',
                self.admin_site.admin_view(self.issue),
                name='django_vox_issue'),
            ] + super().get_urls()

    def preview(self, request, backend_id):
        if not request.user.is_staff:
            raise PermissionDenied
        if request.method != 'POST':
            return django.http.HttpResponseNotAllowed(('POST',))

        try:
            backend = registry.backends.by_id(backend_id)
            message = request.POST['body']
            params = {}
            result = backend.preview_message('', message, params)
        except Exception as exc:
            result = 'Unable to make preview: {}'.format(str(exc))
        return django.http.HttpResponse(result)

    def notification_preview(self, request, notification_id, backend_id):
        if not self.has_change_permission(request):
            raise PermissionDenied
        notification = self.get_object(request, unquote(notification_id))
        if notification is None:
            raise django.http.Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {'name': force_text(self.model._meta.verbose_name),
                   'key': escape(notification_id)})
        if request.method != 'POST':
            return django.http.HttpResponseNotAllowed(('POST',))
        try:
            result = notification.preview(backend_id, request.POST['body'])
        except Exception as exc:
            result = 'Unable to make preview: {}'.format(str(exc))
        return django.http.HttpResponse(result)

    def variables(self, request, notification_id):
        if not self.has_change_permission(request):
            raise PermissionDenied
        notification = self.get_object(request, unquote(notification_id))
        if notification is None:
            raise django.http.Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {'name': force_text(self.model._meta.verbose_name),
                   'key': escape(notification_id)})
        if request.method != 'POST':
            return django.http.HttpResponseNotAllowed(('POST',))
        result = notification.get_recipient_variables()
        return django.http.JsonResponse(result, safe=False)

    def issue(self, request, notification_id, form_url=''):
        if not self.has_change_permission(request):
            raise PermissionDenied
        notification = self.get_object(request, unquote(notification_id))
        if notification is None:
            raise django.http.Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {'name': force_text(self.model._meta.verbose_name),
                   'key': escape(notification_id)})
        if request.method == 'POST':
            form = self.issue_form(notification, request.POST)
            if form.is_valid():
                form.save()
                msg = gettext('Notification sent successfully.')
                django.contrib.messages.success(
                    request, msg, fail_silently=True)
                return django.http.HttpResponseRedirect(
                    reverse(
                        '%s:%s_%s_change' % (
                            self.admin_site.name,
                            notification._meta.app_label,
                            notification._meta.model_name,
                        ),
                        args=(notification.pk,),
                    )
                )
        elif request.method == 'GET':
            form = self.issue_form(notification)
        else:
            return django.http.HttpResponseNotAllowed(('POST',))

        fieldsets = [(None, {'fields': list(form.base_fields)})]
        admin_form = admin.helpers.AdminForm(form, fieldsets, {})

        context = {
            'title': _('Send notification: %s') % escape(str(notification)),
            'media': self.media + admin_form.media,
            'form_url': form_url,
            'form': form,
            'is_popup': (IS_POPUP_VAR in request.POST or
                         IS_POPUP_VAR in request.GET),
            'add': True,
            'change': False,
            'has_delete_permission': False,
            'has_change_permission': True,
            'has_absolute_url': False,
            'opts': self.model._meta,
            'original': notification,
            'save_as': False,
            'show_save': True,
        }
        context.update(self.admin_site.each_context(request))
        request.current_app = self.admin_site.name

        status = 200 if form.is_valid() else 400
        return TemplateResponse(request, self.issue_template, context,
                                status=status)

    def get_readonly_fields(self, request, obj=None):
        if self.has_delete_permission(request, obj):
            return ['object_type'] if obj else []
        return ['codename', 'object_type', 'description', 'required',
                'actor_type', 'target_type']

    def has_delete_permission(self, request, obj=None):
        if obj is None:
            return super().has_delete_permission(request)
        return super().has_delete_permission(request) and not obj.from_code

    def save_form(self, request, form, change):
        result = super().save_form(request, form, change)
        if not change:
            result.from_code = False
        return result

    @staticmethod
    def template_count(obj):
        return obj.template_set.count()


class NotifyForm(forms.Form):

    backend = BackendChoiceField(label=_('Backend'))
    from_address = forms.CharField(label=_('From Address'), required=False)
    subject = forms.CharField(label=_('Subject'), required=False)
    content = forms.CharField(label=_('Content'), required=True,
                              widget=forms.Textarea)

    def __init__(self, protocols, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['content'].widget.attrs['data-preview-url'] = reverse(
            'admin:django_vox_preview', args=('__backend__',))
        backends = [b for bs in (registry.backends.by_protocol(p)
                                 for p in protocols)
                    for b in bs]
        self.fields['backend'].set_backend_choices(backends)


def notify(modeladmin, request, queryset):
    model = getattr(modeladmin, 'model')
    opts = getattr(model, '_meta')
    if not opts:
        raise RuntimeError('modeladmin must be a django ModelAdmin')

    notification = models.OneTimeNotification
    # we don't want to have to evaluate this too many times
    query_list = list(queryset)
    protocols = set()
    for obj in query_list:
        for contact in obj.get_contacts_for_notification(notification):
            protocols.add(contact.protocol)

    if request.POST.get('post'):
        form = NotifyForm(protocols, request.POST)
    else:
        form = NotifyForm(protocols)
    context = {
        'title': _('Notify'),
        'opts': opts,
        'form': form,
        'action_checkbox_name': ACTION_CHECKBOX_NAME,
        'queryset': query_list,
        'fieldset': Fieldset(
            form, fields=('backend', 'from_address', 'subject', 'content')),
    }

    if request.POST.get('post') == 'yes':
        if form.is_valid():
            # okay, now we issue the notification
            contacts = [c for cs in
                        (obj.get_contacts_for_notification(notification)
                         for obj in query_list)
                        for c in cs]
            result = notification.send(
                form.cleaned_data['backend'], contacts,
                form.cleaned_data['from_address'],
                form.cleaned_data['subject'],
                form.cleaned_data['content'])
            if result is None:
                msg = gettext('Notification sent successfully.')
                django.contrib.messages.success(
                    request, msg, fail_silently=True)
            else:
                msg = gettext('Error sending notification (%s).') % result
                django.contrib.messages.error(
                    request, msg, fail_silently=True)
            return django.http.HttpResponseRedirect(
                reverse(
                    '%s:%s_%s_changelist' % (
                        modeladmin.admin_site.name,
                        opts.app_label,
                        opts.model_name,
                    )))

    return TemplateResponse(request, 'django_vox/notify.html', context)


class SiteContactSettingInline(admin.TabularInline):
    model = models.SiteContactSetting


class SiteContactAdmin(admin.ModelAdmin):
    form = SiteContactForm
    inlines = [SiteContactSettingInline]
    list_display = ('name', 'address', 'protocol')


def resend(_admin, _request, queryset):
    for failed in queryset.all():
        failed.resend()


class FailedMessageAdmin(admin.ModelAdmin):
    list_display = ('backend', 'from_address', 'to_addresses', 'created_at')
    list_filter = ('backend', 'from_address', 'to_addresses')
    actions = (resend,)

    # there's no real point in an add button here
    def has_add_permission(self, request):
        return False


admin.site.register(models.SiteContact, SiteContactAdmin)
admin.site.register(models.FailedMessage, FailedMessageAdmin)
admin.site.register(models.Notification, NotificationAdmin)




from django.db import migrations, models


def set_not_bulk(apps, _schema_editor):
    template_cls = apps.get_model('django_vox', 'Template')
    template_cls.objects.update(bulk=False)


class Migration(migrations.Migration):

    dependencies = [
        ('django_vox', '0006_activity_read_at'),
    ]

    operations = [
        migrations.RemoveField(model_name='failedmessage',
                               name='contact_name'),
        migrations.RenameField(model_name='failedmessage',
                               old_name='address', new_name='to_addresses'),
        migrations.AddField(
            model_name='failedmessage', name='from_address',
            field=models.CharField(blank=True, max_length=500,
                                   verbose_name='from address'),
        ),
        migrations.AlterField(
            model_name='failedmessage', name='to_addresses',
            field=models.TextField(verbose_name='to addresses'),
        ),

        migrations.RenameField(model_name='template',
                               old_name='recipient', new_name='recipients'),
        migrations.AddField(
            model_name='template', name='from_address',
            field=models.CharField(blank=True, max_length=500,
                                   verbose_name='from address'),
        ),
        migrations.AddField(
            model_name='template', name='bulk',
            field=models.BooleanField(
                default=True,
                verbose_name='bulk',
                help_text='Send the same message to all recipients'),
        ),
        migrations.AlterField(
            model_name='template', name='recipients',
            field=models.TextField(
                default='re', verbose_name='recipients', blank=True,
                help_text='Who this message actually gets sent to.'),
        ),
        migrations.RunPython(set_not_bulk, migrations.RunPython.noop),
    ]



import json

from django.test import TestCase

import django_vox.backends.template_email


class TestTemplateEmailBackend(TestCase):

    TEXT = '{% block text_body%}' \
           'Here is a message \n\n for {{ you }}' \
           '{% endblock %}' \
           '{% block html_body %}' \
           '<p>Here is a message <br/><br/> for {{ you }}' \
           '{% endblock %}' \
                   ''
    PARAMS = {'you': 'me'}
    SUBJECT = 'SUBJECT'
    MESSAGE = {'subject': 'SUBJECT',
               'text': 'Here is a message \n\n for me',
               'html': '<p>Here is a message <br/><br/> for me'}
    PREVIEW = '<p>Here is a message <br/><br/> for me'

    @classmethod
    def test_build_message(cls):
        backend = django_vox.backends.template_email.Backend
        message = backend.build_message(cls.SUBJECT, cls.TEXT, cls.PARAMS, [])
        obj = json.loads(message)
        assert cls.MESSAGE == obj

    @classmethod
    def test_preview_message(cls):
        backend = django_vox.backends.template_email.Backend
        message = backend.preview_message(cls.SUBJECT, cls.TEXT, cls.PARAMS)
        assert cls.PREVIEW == message



"""
URL patterns for activitypub, some of them aren't implemented
"""
from django.conf.urls import include, url

from . import views

sub_patterns = [
    url(r'^following$', views.empty),
    url(r'^followers$', views.empty),
    url(r'^liked$', views.empty),
    url(r'^inbox$', views.inbox),
    url(r'^outbox$', views.outbox),
    url(r'^$', views.endpoint),
]
urlpatterns = [
    url(r'^(?P<path>.*)/', include(sub_patterns)),
]



import django.template
from django.utils.translation import ugettext_lazy as _

from . import base, base_email

__all__ = ('Backend',)


class Backend(base_email.Backend):

    ID = 'email'
    EDITOR_TYPE = 'html'
    VERBOSE_NAME = _('Email (Template-based)')

    @classmethod
    def build_multipart(cls, subject: str, body: str, parameters: dict):
        return email_parts(subject, body, parameters)


# inspired by django-templated-mail
def email_parts(subject: str, body: str, parameters: dict) -> \
        base_email.MultipartMessage:
    message = base_email.MultipartMessage()
    html_context = django.template.Context(parameters)
    text_context = django.template.Context(parameters, autoescape=False)
    node_parts = {}
    subject_template = base.template_from_string(subject)
    body_template = base.template_from_string(body)
    message.subject = subject_template.render(text_context)

    for node in body_template.nodelist:
        name = getattr(node, 'name', None)
        if name is not None:
            node_parts[name] = node

    with html_context.bind_template(body_template):
        with text_context.bind_template(body_template):
            if 'subject' in node_parts:
                message.subject = node_parts['subject'].render(
                    text_context).strip()
            has_parts = False
            if 'text_body' in node_parts:
                has_parts = True
                message.text = node_parts['text_body'].render(
                    text_context).strip()
            if 'html_body' in node_parts:
                has_parts = True
                message.html = node_parts['html_body'].render(
                    html_context).strip()
            if not has_parts:
                message.text = body_template.render(text_context).strip()
    return message



from unittest import mock

import django.http
from django.contrib.admin import AdminSite
from django.core.exceptions import PermissionDenied
from django.test import TestCase

import django_vox.admin
import django_vox.models

from . import models


class MockAnonUser:

    is_active = False
    is_staff = False
    is_authenticated = False

    @staticmethod
    def has_perm(_perm):
        return False


class MockSuperUser:

    is_active = True
    is_staff = True
    is_authenticated = True

    @staticmethod
    def has_perm(_perm):
        return True


class MockRequest:

    def __init__(self, method, parameters=None):
        self.method = method
        self.user = MockSuperUser()
        self.GET = {}
        self.POST = {}
        if parameters is not None:
            if method == 'POST':
                self.POST = parameters
            else:
                self.GET = parameters
        self.META = {'SCRIPT_NAME': ''}


class MockAnonRequest(MockRequest):
    def __init__(self, method, parameters=None):
        super().__init__(method, parameters)
        self.user = MockAnonUser()


class VariableTests(TestCase):
    """Test the variables that are used in admin"""

    fixtures = ['test']

    @staticmethod
    def test_variables():
        notification = models.Article.get_notification('create')
        variables = notification.get_recipient_variables()
        assert {
                   'si', 'se', 'c:sub', 'c:author', '_static',
                   're:sub', 're:author',
               } == variables.keys()
        # check recipient variables first:
        for key in ('si', 'se', 'c:sub', 'c:author'):
            assert variables[key]['value'] == 'recipient'
        static = variables['_static']
        assert len(static) == 3
        assert static[0]['label'] == 'Article'
        assert static[0]['value'] == 'object'
        assert static[1]['label'] == 'Actor'
        assert static[1]['value'] == 'actor'
        assert static[2]['label'] == 'Target'
        assert static[2]['value'] == 'target'


class TemplateFormTests(TestCase):
    """
    Test template specific forms
    """

    fixtures = ['test']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_class = django_vox.models.Template
        self.form_class = django_vox.admin.TemplateForm

    def test_template_form(self):
        template = self.model_class.objects.get(pk=1)
        notification = template.notification
        blank_form = self.form_class()
        add_form = self.form_class(notification=notification)
        change_form = self.form_class(
            notification=notification, instance=template)

        add_initial = {'attachments': [], 'recipients': ['re']}
        change_initial = {
            'notification': 1,
            'backend': 'email-html',
            'subject': 'Subscriber email',
            'content': ...,
            'recipients': ['c:sub'],
            'from_address': '',
            'enabled': True,
            'bulk': False,
            'attachments': ['object.author.vcard'],
        }

        assert add_initial == add_form.initial
        for key, value in change_initial.items():
            if value is not ...:
                assert value == change_form.initial.get(key)

        assert [] == blank_form.fields['recipients'].choices
        assert [] == blank_form.fields['attachments'].choices
        re_choices = [('si', 'Site Contacts'),
                      ('re:sub', "Target's Subscribers"),
                      ('re:author', "Target's Author"),
                      ('se', 'Actor'),
                      ('c:sub', 'Subscribers'),
                      ('c:author', 'Author')]
        assert re_choices == add_form.fields['recipients'].choices
        at_choices = [('target.author.vcard', 'Target/Author/Contact Info'),
                      ('actor.vcard', 'Actor/Contact Info'),
                      ('object.author.vcard', 'Author/Contact Info')]
        assert at_choices == add_form.fields['attachments'].choices

        backend_opts = add_form.fields['backend'].widget.optgroups(
            'backend', 'email-html')
        opt = next(x[0] for _, x, _idx in backend_opts
                   if x[0]['value'] == 'email-html')

        opt_attrs = {'data-editor': 'html',
                     'data-subject': 'true',
                     'data-from_address': 'true',
                     'data-attachment': 'true',
                     'selected': True}
        assert 'Email (HTML)' == opt['label']
        assert opt['selected']
        assert opt_attrs == opt['attrs']

    def test_template_form_save(self):
        template = self.model_class.objects.get(pk=1)
        data = {
            'attachments': ['actor.vcard', 'object.author.vcard'],
            'notification': 1,
            'backend': 'email-html',
            'content': 'bad content {% if %}',
        }
        change_form = self.form_class(
            data=data, notification=template.notification, instance=template)
        change_form.full_clean()
        assert not change_form.is_valid()
        data['content'] = 'okay content'
        change_form.full_clean()
        assert change_form.is_valid()
        change_form.save()

        new_attachments = self.model_class.objects.get(pk=1).attachments.all()
        new_attachment_keys = {'actor.vcard', 'object.author.vcard'}
        assert new_attachment_keys == set((a.key for a in new_attachments))


class NotificationAdminTests(TestCase):
    """Test NotificationAdmin"""

    fixtures = ['test']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_class = django_vox.models.Notification
        self.admin = django_vox.admin.NotificationAdmin(
            self.model_class, AdminSite())

    def test_template_count(self):
        model_obj = self.model_class.objects.get_by_natural_key(
            'tests', 'article', 'create')
        assert 2 == self.admin.template_count(model_obj)

    def test_get_urls(self):
        names = (u.name for u in self.admin.get_urls() if u.name is not None)
        name_set = set((un for un in names
                        if not un.startswith('django_vox_notification')))
        assert 3 == len(name_set)
        expected = {'django_vox_issue', 'django_vox_preview',
                    'django_vox_variables'}
        assert expected == name_set

    def test_fields(self):
        base_fields = ['codename', 'object_type', 'description']
        from_code_ro = ['codename', 'object_type', 'description',
                        'required', 'actor_type', 'target_type']
        new_obj = self.model_class()
        model_obj = self.model_class.objects.get_by_natural_key(
            'tests', 'article', 'create')
        request = MockRequest('GET')
        assert base_fields == self.admin.get_fields(request)
        assert [] == self.admin.get_readonly_fields(request)
        assert ['object_type'] == self.admin.get_readonly_fields(
            request, new_obj)
        assert from_code_ro == self.admin.get_readonly_fields(
            request, model_obj)

    def test_issue(self):
        notification = self.model_class.objects.get_by_natural_key(
            'tests', 'article', 'create')
        obj = models.Article.objects.first()
        with mock.patch('django_vox.models.Notification.issue'):
            # no POST parameters, so won't be valid
            request = MockRequest('POST', {})
            response = self.admin.issue(request, str(notification.id))
            assert response.status_code == 400
            assert 'This field is required' in response.rendered_content
            # this should 404
            with self.assertRaises(django.http.Http404):
                self.admin.issue(request, '9000')
            # this POST should work
            # we have to perform some malarky because of an isinstance
            # check in django 1.10
            request = MockRequest('POST', {'objects': (obj.pk,)})
            mock_request = mock.Mock(spec=django.http.HttpRequest)
            for attr in request.__dir__():
                if not attr.startswith('_'):
                    setattr(mock_request, attr, getattr(request, attr))
            response = self.admin.issue(mock_request, str(notification.id))
            assert response.status_code == 302

    def test_get_inline_instances(self):
        request = MockRequest('GET', {})
        obj = self.model_class.objects.get_by_natural_key(
            'tests', 'article', 'create')
        assert [] == self.admin.get_inline_instances(request)
        inline = self.admin.get_inline_instances(request, obj=obj)[0]
        formset = inline.get_formset(request, obj=obj)(instance=obj)
        assert obj == formset.get_form_kwargs(0)['notification']

    def test_preview(self):
        anon_request = MockAnonRequest('POST', {'body': 'Hello World'})
        get_request = MockRequest('GET', {'body': 'Hello World'})
        post_request = MockRequest('POST', {'body': 'Hello World'})

        with self.assertRaises(PermissionDenied):
            self.admin.preview(anon_request, 'email-html')

        response = self.admin.preview(get_request, 'email-html')
        assert response.status_code == 405

        response = self.admin.preview(post_request, 'email-html9000')
        assert response.status_code == 200
        assert b"Unable to make preview: 'email-html9000'" == response.content

        response = self.admin.preview(post_request, 'email-html')
        assert response.status_code == 200
        assert b'Hello World' == response.content

    def test_notification_preview(self):
        anon_request = MockAnonRequest('POST', {'body': 'Hello World'})
        get_request = MockRequest('GET', {'body': 'Hello World'})
        syntax_request = MockRequest('POST', {'body': '{% if %}'})
        good_request = MockRequest('POST', {'body': 'Hello {{content}}'})
        backend = 'email-html'
        bad_backend = 'email-html9000'
        notification = '1'
        bad_notification = '9000'

        with self.assertRaises(PermissionDenied):
            self.admin.notification_preview(
                anon_request, notification, backend)

        with self.assertRaises(django.http.Http404):
            self.admin.notification_preview(
                good_request, bad_notification, backend)

        response = self.admin.notification_preview(
            good_request, notification, bad_backend)
        assert 200 == response.status_code
        assert b"Unable to make preview: 'email-html9000'" == response.content

        response = self.admin.notification_preview(
            get_request, notification, backend)
        assert response.status_code == 405

        response = self.admin.notification_preview(
            syntax_request, notification, backend)
        assert response.status_code == 200
        assert response.content.startswith(b'Unable to make preview:')

        response = self.admin.notification_preview(
            good_request, notification, backend)
        assert response.status_code == 200
        assert b'Hello Why are there so many blog demos' == response.content


def test_notify_form():
    protocols = ('email',)
    form = django_vox.admin.NotifyForm(protocols)
    backend_ids = {'email-md', 'email-html', 'postmark-template', 'email'}
    preview_url = '/admin/django_vox/notification/preview/__backend__/'
    assert preview_url == form.fields[
        'content'].widget.attrs['data-preview-url']
    assert backend_ids == set((key for key, _ in
                               form.fields['backend'].choices))



import base64
import json
from typing import List

import django.conf
import django.core.mail.backends.base
import django.core.mail.backends.smtp
import django.template
import django.utils.html
from django.core import mail
from django.utils.text import re_newlines
from django.utils.translation import ugettext_lazy as _

from . import base


class MultipartMessage:

    def __init__(self):
        self.subject = ''
        self.text = ''
        self.html = ''
        self.attachments = []

    @classmethod
    def from_dict(cls, obj):
        result = cls()
        result.subject = obj.get('subject')
        result.text = obj.get('text')
        result.html = obj.get('html')
        result.attachments = [
            base.AttachmentData(
                base64.b64decode(a['data'].encode()),
                a['mime']) for a in obj.get('attachments', ())
        ]
        return result

    @classmethod
    def from_string(cls, string):
        return cls.from_dict(json.loads(string))

    def to_dict(self):
        result = {
            'subject': self.subject,
            'text': self.text,
            'html': self.html,
        }
        if self.attachments:
            result['attachments'] = [
                {'data': base64.b64encode(a.data).decode(), 'mime': a.mime}
                for a in self.attachments]
        return result

    def __str__(self):
        return json.dumps(self.to_dict())

    def to_mail(self) -> mail.EmailMultiAlternatives:
        email = mail.EmailMultiAlternatives()
        email.subject = re_newlines.sub(' ', self.subject)
        if self.text:
            email.body = self.text
            if self.html:
                email.attach_alternative(self.html, 'text/html')
        elif self.html:
            email.body = self.html
            email.content_subtype = 'html'
        for attachment in self.attachments:
            # workaround for django 1.10 problem
            data, mime = attachment.data, attachment.mime
            if mime.startswith('text/'):
                data = data.decode()
            email.attach(content=data, mimetype=mime)
        return email

    def attach(self, data: base.AttachmentData):
        self.attachments.append(data)


class Backend(base.Backend):

    ID = 'email-basic'
    PROTOCOL = 'email'
    VERBOSE_NAME = _('Email (Basic)')
    USE_SUBJECT = True
    USE_ATTACHMENTS = True
    USE_FROM_ADDRESS = True

    @classmethod
    def build_multipart(cls, subject: str, body: str,
                        parameters: dict) -> MultipartMessage:
        raise NotImplementedError()

    @classmethod
    def build_message(cls, subject: str, body: str, parameters: dict,
                      attachments: List[base.AttachmentData]):
        mpm = cls.build_multipart(subject, body, parameters)
        for attachment in attachments:
            mpm.attach(attachment)
        return str(mpm)

    @classmethod
    def preview_message(cls, subject: str, body: str, parameters: dict):
        parts = cls.build_multipart(subject, body, parameters)
        if parts.html:
            return parts.html
        return django.utils.html.escape(parts.text)

    def send_message(self, from_address, to_addresses, message):
        mpm = MultipartMessage.from_string(message)
        email = mpm.to_mail()
        email.from_email = from_address
        email.to = list(to_addresses)
        connection = django.core.mail.get_connection()
        connection.send_messages([email])

    @staticmethod
    def get_default_from_address():
        return django.conf.settings.DEFAULT_FROM_EMAIL



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-10-25 11:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_vox', '0005_activity_table'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='inboxitem',
            name='read',
        ),
        migrations.AddField(
            model_name='inboxitem',
            name='read_at',
            field=models.DateTimeField(
                blank=True, db_index=True, null=True,
                verbose_name='date read'),
        ),
    ]



import json

import aspy
from django.test import Client, TestCase

import django_vox.models
import django_vox.settings
from tests import models

EXTRA = {'HTTP_ACCEPT': 'application/activity+json'}


class WebTests(TestCase):
    """Test the walkthrough in the documentation"""

    fixtures = ['test']

    @staticmethod
    def test_bad_requests():
        client = Client()
        get_list = (
            ('/~1/inbox', 200, "Sanity check failed"),
            ('/~404/', 404, "Didn't return 404 on bad endpoint"),
            ('/~404/inbox', 404, "Didn't return 404 on bad inbox"),
            ('/~2/inbox', 403, "Didn't return 403 on other user's inbox"),
            ('/~1/outbox', 405, "Didn't return 405 on getting outbox"),
        )
        post_list = (
            ('/~404/outbox', 404, "Didn't return 404 on bad outbox"),
            ('/~2/outbox', 403, "Didn't return 403 on other user's outbox"),
            ('/~1/inbox', 405, "Didn't return 405 on getting inbox"),
        )
        response = client.post('/admin/login/',
                               {'username': 'author@example.org',
                                'password': 'password'})
        assert 302 == response.status_code, 'login failed'
        response = client.get('/~1/inbox', **EXTRA)
        for url, status, info in get_list:
            response = client.get(url, **EXTRA)
            assert status == response.status_code, info
        for url, status, info in post_list:
            response = client.post(url, {}, **EXTRA)
            assert status == response.status_code, info
        # test and empty endpoint
        response = client.get('/~1/followers', **EXTRA)
        assert 200 == response.status_code
        data = json.loads(response.content)
        assert 'Collection' == data['type']
        assert 'items' not in data

    @staticmethod
    def test_actor_page():
        client = Client()
        response = client.get('/~1/', **EXTRA)
        assert 200 == response.status_code
        assert 'application/activity+json' == response['Content-Type']
        json_obj = json.loads(response.content)
        assert 'Person' == json_obj['type']
        assert 'http://127.0.0.1:8000/~1/' == json_obj['id']

    @staticmethod
    def test_article_add():
        # sanity check
        assert django_vox.models.InboxItem.objects.count() == 0
        article = models.Article.objects.get(pk='too_many')
        subscriber = models.Subscriber.objects.get(id=1)
        author_subscriber = models.Subscriber.objects.get(id=2)
        # first we create an article as the author user
        models.Comment.objects.create(
            article=article,
            poster=subscriber,
            content='First Post!11',
        )
        models.Comment.objects.create(
            article=article,
            poster=author_subscriber,
            content='This is why we can\'t have nice things',
        )
        # now two notification should have been fired, check the
        # inbox items
        assert 2 == django_vox.models.InboxItem.objects.count()
        # now, let's check the results of the inbox endpoint
        client = Client()

        # fist we have to authenticate
        response = client.post('/admin/login/',
                               {'username': 'author@example.org',
                                'password': 'password'})
        assert 302 == response.status_code, 'login failed'
        response = client.get('/~1/inbox', **EXTRA)
        assert 200 == response.status_code
        assert 'application/activity+json' == response['Content-Type']
        json_obj = json.loads(response.content)
        assert 'OrderedCollection' == json_obj['type']
        items = json_obj['items']
        assert 2 == len(items)
        # in make sure this is in descending order
        assert 'http://127.0.0.1:8000/too_many/#comment-1' == \
               items[1]['object']['id']
        assert 'First Post!11' == items[1]['object']['content']
        assert 'Note' == items[1]['object']['name']
        assert 'http://127.0.0.1:8000/too_many/#comment-2' == \
               items[0]['object']['id']
        assert 'Author Subscriber' == items[0]['actor']['name']

    @staticmethod
    def test_activity_read():
        # sanity check
        assert django_vox.models.InboxItem.objects.count() == 0
        article = models.Article.objects.get(pk='too_many')
        subscriber = models.Subscriber.objects.get(id=1)
        # first we create an article as the author user
        models.Comment.objects.create(
            article=article,
            poster=subscriber,
            content='read me',
        )
        # now one notification should have been fired, check the
        # inbox items
        assert 1 == django_vox.models.InboxItem.objects.count()
        # now we'll check the activity
        client = Client()
        response = client.post('/admin/login/',
                               {'username': 'author@example.org',
                                'password': 'password'})
        assert 302 == response.status_code, 'login failed'
        # make sure inbox looks normal
        response = client.get('/~1/inbox', **EXTRA)
        assert 200 == response.status_code
        assert 'application/activity+json' == response['Content-Type']
        json_obj = json.loads(response.content)
        assert 'OrderedCollection' == json_obj['type']
        items = json_obj['items']
        assert 1 == len(items)
        item_id = items[0]['id']
        # check db
        inbox = django_vox.models.InboxItem.objects.all()
        assert 1 == len(inbox)
        assert inbox[0].read_at is None
        # now mark the message as read
        activity = aspy.Read(object=item_id)
        response = client.post('/~1/outbox', activity.to_dict(), **EXTRA)
        assert 200 == response.status_code
        # now we should have one read item
        inbox = django_vox.models.InboxItem.objects.all()
        assert 1 == len(inbox)
        assert inbox[0].read_at is not None



import urllib.parse

import aspy
import django.http
import django.shortcuts

from . import models, registry, settings


def get_local_urlpart(full_uri):
    _, _, path, query, fragment = urllib.parse.urlsplit(full_uri)
    return urllib.parse.urlunsplit(('', '', path, query, fragment))


def fix_path(func):
    """
    Django likes to strip the trailing slash, but we need it for
    further matching
    """
    def inner(request, path):
        path += '/'
        return func(request, path)

    return inner


@fix_path
def endpoint(_request, path):
    try:
        actor = registry.objects.get_local_object(path)
    except registry.ObjectNotFound:
        return django.http.HttpResponseNotFound()
    obj = actor.__activity__()
    obj_id = obj['id']
    for part in ('followers', 'following', 'liked', 'inbox', 'outbox'):
        if part not in obj:
            obj[part] = urllib.parse.urljoin(obj_id, part)

    return django.http.HttpResponse(
        str(obj), content_type='application/activity+json')


@fix_path
def outbox(request, path):
    try:
        owner = registry.objects.get_local_object(path)
    except registry.ObjectNotFound:
        return django.http.HttpResponseNotFound()
    if request.user != owner:
        return django.http.HttpResponseForbidden()
    elif request.method == 'POST':
        return outbox_post(request, owner)
    else:
        return django.http.HttpResponseNotAllowed(
            permitted_methods=('POST',))


@fix_path
def inbox(request, path):
    try:
        owner = registry.objects.get_local_object(path)
    except registry.ObjectNotFound:
        return django.http.HttpResponseNotFound()
    if request.user != owner:
        return django.http.HttpResponseForbidden()
    if request.method == 'GET':
        return inbox_get(request, owner)
    else:
        return django.http.HttpResponseNotAllowed(
            permitted_methods=('GET',))


def inbox_get(_request, owner):
    query = models.InboxItem.objects.select_related('activity').filter(
        owner=owner).order_by('-id')
    items = []
    for record in query[:settings.INBOX_LIMIT]:
        item = record.activity.__activity__()
        if record.read_at:
            item['https://schema.org/dateRead'] = record.read_at
        items.append(item)

    collection = aspy.OrderedCollection(
        summary='Inbox for {}'.format(owner),
        items=items)
    return django.http.HttpResponse(
        str(collection), content_type='application/activity+json')


def outbox_post(request, owner):
    obj = request.POST.get('object')
    activity_type = request.POST.get('type')
    if not obj:
        return django.http.HttpResponseBadRequest('Object field required')
    if not activity_type:
        return django.http.HttpResponseBadRequest('Type field required')
    if isinstance(obj, str):
        iri = obj
    elif isinstance(obj, dict):
        iri = obj.get('id')
    else:
        return django.http.HttpResponseBadRequest(
            'Unrecognized format for object')
    if not iri:
        return django.http.HttpResponseBadRequest('Object missing ID')
    # now we have to strip our domain
    local_iri = get_local_urlpart(iri)
    if activity_type == 'Create':
        local_object = registry.objects.create_local_object(local_iri)
    else:
        local_object = registry.objects.get_local_object(local_iri)
    func = 'activity_{}'.format(activity_type.lower())
    method = getattr(local_object, func)
    if method and callable(method):
        method(request.POST, owner)
        return django.http.HttpResponse(status=200, content='')


@fix_path
def empty(_request, _path):
    return django.http.HttpResponse(
        str(aspy.Collection()), content_type='application/activity+json')



import json

from django.test import TestCase
import django.conf
import django.core.mail

from django_vox.backends.html_email import Backend


class TestHtmlEmailBackend(TestCase):

    TEXT = '<p>Here is a message <br/><br/> for {{ you }}'
    PARAMS = {'you': 'me'}
    SUBJECT = 'SUBJECT'
    MESSAGE = {'subject': 'SUBJECT',
               'text': 'Here is a message\n\nfor me',
               'html': '<p>Here is a message <br/><br/> for me'}
    PREVIEW = '<p>Here is a message <br/><br/> for me'

    @classmethod
    def test_build_message(cls):
        message = Backend.build_message(cls.SUBJECT, cls.TEXT, cls.PARAMS, [])
        obj = json.loads(message)
        assert cls.MESSAGE == obj

    @classmethod
    def test_preview_message(cls):
        message = Backend.preview_message(cls.SUBJECT, cls.TEXT, cls.PARAMS)
        assert cls.PREVIEW == message

    def test_send_message(self):
        message = Backend.build_message(
            self.SUBJECT, self.TEXT, self.PARAMS, [])

        from_address = django.conf.settings.DEFAULT_FROM_EMAIL
        to_addresses = ['george@example.test']

        instance = Backend()
        instance.send_message(from_address, to_addresses, message)
        assert 1 == len(django.core.mail.outbox)
        mail = django.core.mail.outbox[0]
        assert mail.from_email == 'admin@example.test'
        assert mail.to == ['george@example.test']
        assert mail.body == self.MESSAGE['text']
        assert mail.alternatives[0][1] == 'text/html'



from django.contrib import admin
import django_vox.admin

from . import models


class UserAdmin(admin.ModelAdmin):
    actions = (django_vox.admin.notify,)


admin.site.register(models.User, UserAdmin)
admin.site.register(models.Article)
admin.site.register(models.Subscriber)
admin.site.register(models.Comment)



"""
URL patterns for testing app.
"""
from django.conf.urls import url

from . import views

app_name = 'tests'
urlpatterns = [
    url(r'^$', views.article_list, name='index'),
    url(r'^~(?P<user_id>[0-9]+)/$', views.user_detail, name='user'),
    url(r'^(?P<article_pk>\w+)/$', views.article_detail, name='article'),
    url(r'^\.(?P<sub_id>[0-9]+)/$', views.subscriber_detail,
        name='subscriber'),
    url(r'^subscribe$', views.subscribe, name='subscribe'),
    url(r'^(?P<article_pk>\w+)/comment$', views.comment,
        name='post-comment'),
]



from django.utils.translation import ugettext_lazy as _

from django_vox import settings

from . import base

__all__ = ('Backend',)


class Backend(base.Backend):
    """This backends implements message sending via XMPP

    Warning, it makes a new connection for each message which in practice can
    be pretty slow.
    """

    ID = 'xmpp'
    PROTOCOL = 'xmpp'
    EDITOR_TYPE = 'basic'
    ESCAPE_HTML = False
    VERBOSE_NAME = _('XMPP')
    DEPENDS = ('sleekxmpp',)

    # TODO: implement __init__ and do the heavy work there
    # so that sending multiple messages at once isn't stupidly slow

    def send_message(self, _from_address, to_addresses, message):
        for address in to_addresses:
            try:
                client = Client(settings.XMPP_JID, settings.XMPP_PASSWORD,
                                address, message)
            except NameError:
                raise ImportError('sleekxmpp is required to use this backend')
            if client.connect():
                client.process(block=True)
            else:
                raise RuntimeError('Unable to connect to XMPP account')


try:
    import sleekxmpp
except ImportError:
    pass
else:
    # largely copied from sleekxmpp examples
    class Client(sleekxmpp.ClientXMPP):
        def __init__(self, jid: str, password: str,
                     recipient: str, message: str):
            super().__init__(jid, password)
            # The message we wish to send, and the JID that will receive it.
            self.recipient = recipient
            self.msg = message
            # The session_start event will be triggered when the bot
            # establishes its connection with the server and the XML
            # streams are ready for use. We want to listen for this
            # event so that we we can initialize our roster.
            self.add_event_handler("session_start", self.start, threaded=True)

        def start(self, _event):
            """
            Process the session_start event.
            Typical actions for the session_start event are
            requesting the roster and broadcasting an initial
            presence stanza.
            Arguments:
                event -- An empty dictionary. The session_start
                         event does not provide any additional
                         data.
            """
            self.send_presence()
            self.get_roster()
            self.send_message(
                mto=self.recipient, mbody=self.msg, mtype='chat')
            # Using wait=True ensures that the send queue will be emptied
            # before ending the session.
            self.disconnect(wait=True)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-10-22 15:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_vox', '0003_activity_stream'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='inboxitem',
            name='json',
        ),
        migrations.AddField(
            model_name='inboxitem',
            name='actor_json',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='inboxitem',
            name='object_json',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='inboxitem',
            name='target_id',
            field=models.CharField(db_index=True, default='', max_length=2048,
                                   verbose_name='actor id'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='inboxitem',
            name='target_json',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='inboxitem',
            name='name',
            field=models.CharField(blank=True, max_length=512,
                                   verbose_name='name'),
        ),
        migrations.AddField(
            model_name='inboxitem',
            name='summary',
            field=models.TextField(blank=True),
        ),
    ]



import datetime
import warnings

import aspy
import pytz
from bs4 import BeautifulSoup
from django.contrib.contenttypes.models import ContentType
from django.core import mail
from django.test import Client, TestCase

import django_vox.models
import django_vox.registry

from . import models


class FailmailTests(TestCase):

    fixtures = ['test']

    def test_email_failure(self):
        assert 0 == len(mail.outbox)
        author = models.User.objects.get(email='author@example.org')
        # by default, with the test config, we throw errors
        with self.settings(EMAIL_BACKEND='Can\'t import this!'):
            with self.assertRaises(ImportError):
                models.Article.objects.create(
                    slug='second',
                    author=author,
                    title='A second article',
                    content='Whoever thought we\'d come this far',
                )
        assert 0 == len(mail.outbox)
        assert 3 == django_vox.models.FailedMessage.objects.count()
        # we can change that though
        with self.settings(
                DJANGO_VOX_THROW_EXCEPTIONS=False,
                EMAIL_BACKEND='Can\'t import this!'):
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                models.Article.objects.create(
                    slug='third',
                    author=author,
                    title='A third article',
                    content='Whoever thought we\'d come this far',
                )
            # this should create 3 failed messages
            assert 0 == len(mail.outbox)
            messages = django_vox.models.FailedMessage.objects.order_by(
                'created_at')
            assert 6 == len(messages)
            # ends with date
            for message in messages:
                assert '@example.org @ ' in str(message)

    def test_email_fail_resend(self):
        with self.settings(
                DJANGO_VOX_THROW_EXCEPTIONS=False,
                EMAIL_BACKEND='Can\'t import this!'):
            assert 0 == len(mail.outbox)
            author = models.User.objects.get(email='author@example.org')
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                models.Article.objects.create(
                    slug='second',
                    author=author,
                    title='A second article',
                    content='Whoever thought we\'d come this far',)
            # this should create 3 failed messages
            messages = django_vox.models.FailedMessage.objects.order_by(
                'created_at')
            assert 3 == len(messages)
            for message in messages:
                assert '@example.org @ ' in str(message)
            # try resending the first one (to admin)
            with self.assertRaises(Exception):
                messages[0].resend()

        messages = django_vox.models.FailedMessage.objects.order_by(
            'created_at')
        messages[0].resend()
        assert 1 == len(mail.outbox)


class ExtraOptionsTest(TestCase):
    """Test extra notification template options"""
    fixtures = ['test']

    @staticmethod
    def test_from_address():
        assert len(mail.outbox) == 0
        # get the article created template
        template = django_vox.models.Template.objects.get(pk=1)
        template.from_address = '{{ object.author.email }}'
        template.save()
        # disable message to site contact
        template = django_vox.models.Template.objects.get(pk=2)
        template.enabled = False
        template.save()
        # first we create an article as the author user
        author = models.User.objects.get(email='author@example.org')
        models.Article.objects.create(
            slug='another',
            author=author,
            title='Just another article',
            content='nothing really matters ... to me',
        )
        assert len(mail.outbox) == 2
        for message in mail.outbox:
            assert message.from_email == 'author@example.org'


class DemoTests(TestCase):
    """Test the walkthrough in the documentation"""

    fixtures = ['test']

    @staticmethod
    def test_walkthrough():
        # sanity
        assert len(mail.outbox) == 0
        # first we create an article as the author user
        author = models.User.objects.get(email='author@example.org')
        models.Article.objects.create(
            slug='second',
            author=author,
            title='A second article',
            content='Whoever thought we\'d come this far',
        )
        # now a notification should be fired, check the outbox
        # mail.outbox is a list of EmailMultiAlternatives
        assert len(mail.outbox) == 3
        for message in mail.outbox:
            # ignore site message
            if 'admin@example.org' in message.to:
                continue
            assert (message.content_subtype == 'plain'
                    and message.mixed_subtype == 'mixed')
            html = message.alternatives[0][0]
            soup = BeautifulSoup(html, 'html.parser')
            anchors = soup.find_all('a')
            assert len(anchors) == 1
            url = anchors[0].get('href')
            assert url.startswith('http://127.0.0.1:8000/second/?token=')
            assert len(url) > 31  # if less, token is blank
            client = Client()
            response = client.get(url)
            assert response.status_code == 200
            assert response['Content-Type'] == 'text/html; charset=utf-8'

    @staticmethod
    def test_markdown():
        # sanity
        assert len(mail.outbox) == 0
        # get the article created template
        template = django_vox.models.Template.objects.get(pk=1)
        template.backend = 'email-md'
        template.subject = 'Hi {{ recipient.name }}'
        template.content = 'Hi {{ recipient.name }},\n\nA new article, ' \
                           '[{{ object }}](http://127.0.0.1:8000/{{ ' \
                           'object.pk }}/?token={{ recipient.get_token | ' \
                           'urlencode }}), has been published at the ' \
                           'awesome blog.'
        template.save()
        # first we create an article as the author user
        author = models.User.objects.get(email='author@example.org')
        models.Article.objects.create(
            slug='second',
            author=author,
            title='A second article',
            content='Whoever thought we\'d come this far',
        )
        # now a notification should be fired, check the outbox
        # mail.outbox is a list of EmailMultiAlternatives
        assert len(mail.outbox) == 3
        mail_by_subject = {}
        for message in mail.outbox:
            mail_by_subject[message.subject] = message
        site_mail = mail_by_subject['Hi Subscriber']
        assert len(site_mail.alternatives) == 1
        soup = BeautifulSoup(site_mail.alternatives[0][0], 'html.parser')
        assert len(site_mail.attachments) == 1
        assert site_mail.attachments[0][2] == 'text/vcard'
        assert 'FN:Author' in site_mail.attachments[0][1]
        anchors = soup.find_all('a')
        assert len(anchors) == 1
        url = anchors[0].get('href')
        assert url.startswith('http://127.0.0.1:8000/second/?token=')
        assert len(url) > 31  # if less, token is blank

    @staticmethod
    def test_choices():
        """
        Test the values of the dropdown fields
        """
        expected_ids = set()
        for model in (models.Article, models.Subscriber, models.User,
                      models.Comment, django_vox.models.SiteContact):
            ct = ContentType.objects.get_for_model(model)
            expected_ids.add(ct.id)

        vox_ct_limit = django_vox.registry.channel_type_limit()
        assert 'id__in' in vox_ct_limit
        actual_ids = set(vox_ct_limit['id__in'])
        assert actual_ids == expected_ids

    @staticmethod
    def test_previews():
        pp = django_vox.models.PreviewParameters(
            models.Article, models.User, models.Subscriber)
        assert 'recipient' not in pp
        assert 'target' in pp
        assert pp['content'].title == 'Why are there so many blog demos'
        assert pp['object'].title == 'Why are there so many blog demos'
        assert pp['target'].name == 'Subscriber'
        assert pp['actor'].name == 'Author'
        # now delete the articles and watch it come up with names
        for article in models.Article.objects.all():
            article.delete()
        for sub in models.Subscriber.objects.all():
            sub.delete()
        pp = django_vox.models.PreviewParameters(
            models.Article, models.User, models.Subscriber)
        assert 'recipient' not in pp
        assert 'target' in pp
        assert pp['object'].title == '{title}'
        assert pp['target'].secret == ''
        assert pp['actor'].name == 'Author'


class TestTemplate(TestCase):

    fixtures = ['test']

    def test_str(self):
        template = django_vox.models.Template.objects.get(pk=1)
        assert 'Email (HTML) for Subscribers' == str(template)


def test_load_aspy_object():
    json = '{\n' \
           '"@context": "https://www.w3.org/ns/activitystreams",\n' \
           '"type": "Create",\n' \
           '"object": {\n' \
           '  "type": "Note",\n' \
           '  "updated": "2018-11-14T06:47:34Z",\n' \
           '  "summary": "Here’s a note"\n' \
           '}\n' \
           '}'
    obj = django_vox.models.load_aspy_object(json)
    assert isinstance(obj['object'], aspy.Note)
    assert obj['object']['updated'] == datetime.datetime(
        2018, 11, 14, 6, 47, 34, tzinfo=pytz.utc)



"""
Django settings for django_vox tests
"""
# based of django-debug-toolbar
# https://github.com/jazzband/django-debug-toolbar/blob/master/tests/settings.py
import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production

SECRET_KEY = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890'

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django_vox',
]
